# Copyright (c) 2024 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Intel Neural Compressor Pytorch quantization base API."""

import copy
from typing import Any, Callable, Dict, Tuple

import torch

from neural_compressor.common.base_config import BaseConfig, ComposableConfig, config_registry
from neural_compressor.common.utils import Mode, call_counter, log_process
from neural_compressor.torch.quantization.config import (
    AutoRoundConfig,
    FP8Config,
    HybridGPTQConfig,
    INT8StaticQuantConfig,
    SmoothQuantConfig,
)
from neural_compressor.torch.utils import is_ipex_available, logger
from neural_compressor.torch.utils.utility import WHITE_MODULE_LIST, algos_mapping, get_model_info

FRAMEWORK_NAME = "torch"


def need_apply(configs_mapping: Dict[Tuple[str, callable], BaseConfig], algo_name):
    """Check whether to apply this algorithm according to configs_mapping.

    Args:
        configs_mapping (Dict[Tuple[str, callable], BaseConfig]): configs mapping
        algo_name (str): algo name

    Returns:
        Bool: True or False.
    """
    return any(config.name == algo_name for config in configs_mapping.values())


def preprocess_quant_config(model, quant_config, mode="prepare", example_inputs=None, run_fn=None):
    """Preprocess the quantization configuration.

    Args:
        model (torch.nn.Module): Float model to be quantized.
        quant_config (BaseConfig | dict): Quantization configuration or configuration dictionary.
        mode (str, optional): Quantization mode to prepare for. Defaults to "prepare".
        example_inputs (Any, optional): Example inputs for tracing the model. Defaults to None.
        run_fn (Callable | None): Calibration function for collecting statistics. Defaults to None.

    Returns:
        tuple[torch.nn.Module, dict]: Updated model and configuration mapping.
    """
    registered_configs = config_registry.get_cls_configs()
    if isinstance(quant_config, dict):
        quant_config = ComposableConfig.from_dict(quant_config, config_registry=registered_configs[FRAMEWORK_NAME])
        logger.info(f"Parsed a config dict to construct the quantization config: {quant_config}.")
    else:
        assert isinstance(
            quant_config, BaseConfig
        ), f"Please pass a dict or config instance as the quantization configuration, but got {type(quant_config)}."
    logger.debug("Quantize model with config:")
    logger.debug(quant_config.to_dict())

    if is_ipex_available and (
        isinstance(quant_config, INT8StaticQuantConfig) or isinstance(quant_config, SmoothQuantConfig)
    ):
        if mode == "quantize":
            if isinstance(quant_config, SmoothQuantConfig):
                from neural_compressor.torch.algorithms.smooth_quant import TorchSmoothQuant

                sq = TorchSmoothQuant(
                    model, dataloader=None, example_inputs=example_inputs, q_func=run_fn, record_max_info=True
                )
                model.sq_info = sq
                model = sq.transform(
                    alpha=quant_config.alpha,
                    folding=quant_config.folding,
                    auto_alpha_args=quant_config.auto_alpha_args,
                    scale_sharing=quant_config.scale_sharing,
                )
        model_info = quant_config.get_model_info(model, example_inputs)
    elif isinstance(quant_config, AutoRoundConfig):
        _tokenizer_backup = getattr(quant_config, "tokenizer", None)
        if _tokenizer_backup is not None:
            setattr(model, "tokenizer", _tokenizer_backup)
            delattr(quant_config, "tokenizer")
        model_info = quant_config.get_model_info(model=model)
    else:
        model_info = quant_config.get_model_info(model=model)

    if (
        (hasattr(quant_config, "model_path") and quant_config.model_path == "")
        or isinstance(quant_config, ComposableConfig)
    ) and hasattr(model, "name_or_path"):
        quant_config.model_path = model.name_or_path
    if hasattr(model, "dpq_quantized") and model.dpq_quantized and type(quant_config) == FP8Config:
        quant_config = HybridGPTQConfig.convert_from_fp8(quant_config)
    configs_mapping = quant_config.to_config_mapping(model_info=model_info)
    logger.debug(configs_mapping)
    return model, configs_mapping


@log_process(mode=Mode.QUANTIZE)
@call_counter
def quantize(
    model: torch.nn.Module,
    quant_config: BaseConfig,
    run_fn: Callable = None,
    run_args: Any = None,
    inplace: bool = True,
    example_inputs: Any = None,
) -> torch.nn.Module:
    """The main entry to quantize model with static mode.

    Args:
        model: a float model to be quantized.
        quant_config: a quantization configuration.
        run_fn: a calibration function for calibrating the model. Defaults to None.
        run_args: positional arguments for `run_fn`. Defaults to None.
        example_inputs: used to trace torch model.

    Returns:
        The quantized model.
    """
    q_model = model if inplace else copy.deepcopy(model)
    q_model, configs_mapping = preprocess_quant_config(
        q_model, quant_config, mode="quantize", example_inputs=example_inputs, run_fn=run_fn
    )
    for algo_name, algo_func in algos_mapping.items():
        # select quantization algo according to config
        if need_apply(configs_mapping, algo_name):
            logger.info(f"Start to apply {algo_name} on the model.")
            q_model = algo_func(
                q_model,
                configs_mapping,
                run_fn=run_fn,
                run_args=run_args,
                example_inputs=example_inputs,
                mode=Mode.QUANTIZE,
            )
    setattr(q_model, "is_quantized", True)
    return q_model


@log_process(mode=Mode.PREPARE)
def prepare(
    model: torch.nn.Module,
    quant_config: BaseConfig,
    inplace: bool = True,
    example_inputs: Any = None,
):
    """Prepare the model for calibration.

    Insert observers into the model so that it can monitor the input and output tensors during calibration.

    Args:
        model (torch.nn.Module): origin model
        quant_config (BaseConfig): path to quantization config
        inplace (bool, optional): It will change the given model in-place if True.
        example_inputs (tensor/tuple/dict, optional): used to trace torch model.

    Returns:
        prepared and calibrated module.
    """
    prepared_model = model if inplace else copy.deepcopy(model)
    prepared_model, configs_mapping = preprocess_quant_config(
        prepared_model, quant_config, mode="prepare", example_inputs=example_inputs
    )
    for algo_name, algo_func in algos_mapping.items():
        # select quantization algo according to config
        if need_apply(configs_mapping, algo_name):
            logger.info(f"Start to prepare model with {algo_name}.")
            prepared_model = algo_func(
                prepared_model,
                configs_mapping,
                example_inputs=example_inputs,
                mode=Mode.PREPARE,
            )
            setattr(prepared_model, "is_prepared", True)
    setattr(prepared_model, "quant_config", quant_config)
    setattr(prepared_model, "example_inputs", example_inputs)
    return prepared_model


@log_process(mode=Mode.PREPARE)
def prepare_qat(
    model: torch.nn.Module,
    mapping=None,
    inplace: bool = True,
):
    r"""Prepares a copy of the model for quantization calibration or
    quantization-aware training and converts it to quantized version.

    Quantization configuration should be assigned preemptively
    to individual submodules in `.qconfig` attribute.

    Args:
        model: input model to be modified in-place
        quant_config: quantization config that maps float modules to quantized modules to be
                 replaced.
        inplace: carry out model transformations in-place, the original module
                 is mutated
    """
    assert model.training, "prepare_qat only works on models in training mode"

    from .config import get_default_qat_module_mappings

    if mapping is None:
        mapping = get_default_qat_module_mappings()

    from ..algorithms.qat.quant_utils import convert_model_with_mapping

    return convert_model_with_mapping(model, mapping)


@log_process(mode=Mode.CONVERT)
def convert(
    model: torch.nn.Module,
    quant_config: BaseConfig = None,
    inplace: bool = True,
    **kwargs,
):
    """Convert the prepared model to a quantized model.

    Args:
        model (torch.nn.Module): torch model
        quant_config (BaseConfig, optional): path to quantization config, only required when model is not prepared.
        inplace (bool, optional): It will change the given model in-place if True.

    Returns:
        The quantized model.
    """
    q_model = model if inplace else copy.deepcopy(model)

    assert (
        getattr(model, "is_prepared", False) or quant_config is not None
    ), "Please pass quant_config to convert function."

    if getattr(model, "is_prepared", False):
        if quant_config is None:
            quant_config = model.quant_config
        else:
            logger.warning("quant_config will be ignored since the model has been prepared.")
    example_inputs = model.example_inputs if getattr(model, "is_prepared", False) else None

    registered_configs = config_registry.get_cls_configs()
    if isinstance(quant_config, dict):
        quant_config = ComposableConfig.from_dict(quant_config, config_registry=registered_configs[FRAMEWORK_NAME])
        logger.info(f"Parsed a config dict to construct the quantization config: {quant_config}.")
    else:
        assert isinstance(
            quant_config, BaseConfig
        ), f"Please pass a dict or config instance as the quantization configuration, but got {type(quant_config)}."
    if hasattr(model, "dpq_quantized") and model.dpq_quantized and type(quant_config) == FP8Config:
        quant_config = HybridGPTQConfig.convert_from_fp8(quant_config)
    logger.debug("Convert model with config:")
    logger.debug(quant_config.to_dict())

    # select quantization algo according to config
    if is_ipex_available and (
        isinstance(quant_config, INT8StaticQuantConfig) or isinstance(quant_config, SmoothQuantConfig)
    ):
        model_info = quant_config.get_model_info(q_model, example_inputs)
    else:
        model_info = quant_config.get_model_info(model=q_model)
    configs_mapping = quant_config.to_config_mapping(model_info=model_info)
    logger.debug(configs_mapping)

    # TODO: Need to consider composableConfig situation
    for algo_name, algo_func in algos_mapping.items():
        if need_apply(configs_mapping, algo_name):
            logger.info(f"Start to convert model with {algo_name}.")
            q_model = algo_func(
                q_model,
                configs_mapping,
                example_inputs=example_inputs,
                mode=Mode.CONVERT,
                **kwargs,
            )
    setattr(q_model, "is_quantized", True)
    return q_model


def finalize_calibration(model):
    """Generate and save calibration info."""
    from neural_compressor.torch.algorithms.fp8_quant import save_calib_result

    save_calib_result(model)



# -*- coding: utf-8 -*-
# Copyright (c) 2025 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


def initialize_model_and_tokenizer(model_name_or_path, use_load=False, device="cpu"):
    import transformers

    from neural_compressor.torch.utils import local_rank, logger, world_size

    tokenizer = transformers.AutoTokenizer.from_pretrained(model_name_or_path)
    if use_load:
        from neural_compressor.torch.quantization import load

        model = load(model_name_or_path, format="huggingface", device=device)
        model, tokenizer = update_tokenizer(model, tokenizer)
        return model, tokenizer
    config = transformers.AutoConfig.from_pretrained(model_name_or_path)
    # using memory mapping with torch_dtype=config.torch_dtype
    model = transformers.AutoModelForCausalLM.from_pretrained(model_name_or_path, torch_dtype=config.torch_dtype)
    model, tokenizer = update_tokenizer(model, tokenizer)
    # shard model for multi-cards and enable hpu graph
    if world_size > 1:
        if "hpu" in device:
            logger.warning("The model will be loaded via memory mapping, so the device settings are useless")
        ds_inference_kwargs = {
            "dtype": config.torch_dtype,
            "tensor_parallel": {"tp_size": world_size},
            "keep_module_on_host": True,
        }
        import deepspeed

        ds_model = deepspeed.init_inference(model, **ds_inference_kwargs)
        model = ds_model.module
    model.eval()
    return model, tokenizer


def update_tokenizer(model, tokenizer):
    if model.config.model_type in ["llama", "mixtral"]:
        # unwind broken decapoda-research config
        model.generation_config.pad_token_id = 0
        model.generation_config.bos_token_id = 1
        model.generation_config.eos_token_id = 2
        tokenizer.bos_token_id = model.generation_config.bos_token_id
        tokenizer.eos_token_id = model.generation_config.eos_token_id
        tokenizer.pad_token_id = model.generation_config.pad_token_id
    return model, tokenizer


def get_default_llm_dataloader(tokenizer, dataset_name="NeelNanda/pile-10k", bs=8, nsamples=128, seq_len=128, seed=42):
    """Generate dataloader based on dataset name and other configurations.

    Args:
        tokenizer (obj): tokenizer object.
        seq_len (int, optional): the sequence length of the input tokens. Defaults to 128.
        dataset_name (str, optional): dataset name. Defaults to "NeelNanda/pile-10k".
        seed (int, optional): random seed. Defaults to 42.
        bs (int, optional): batch size. Defaults to 8.
        nsamples (int, optional): number of samples. Defaults to 128.

    Returns:
        dataloader: dataloader
    """
    from datasets import load_dataset
    from torch.utils.data import DataLoader, Dataset

    dataset = load_dataset(dataset_name, split="train")
    dataset = dataset.shuffle(seed=seed).select(range(nsamples))

    class TokenizedDataset(Dataset):
        def __init__(self, dataset, tokenizer, seq_len):
            self.dataset = dataset
            self.tokenizer = tokenizer
            self.seq_len = seq_len

        def __len__(self):
            return len(self.dataset)

        def __getitem__(self, idx):
            text = self.dataset[idx]["text"]
            inputs = self.tokenizer(
                text, max_length=self.seq_len, padding="max_length", truncation=True, return_tensors="pt"
            )
            return {key: val.squeeze(0) for key, val in inputs.items()}

    tokenized_dataset = TokenizedDataset(dataset, tokenizer, seq_len)
    dataloader = DataLoader(tokenized_dataset, batch_size=bs, shuffle=True)
    return dataloader


def llm_benchmark(model, batch_size, input_length, warmup_iters=3, total_iters=20):
    import time

    import torch

    from neural_compressor.torch.utils import get_accelerator, logger

    cur_accelerator = get_accelerator()
    # this is a simple example to show the performance benefit of quantization
    example_inputs = torch.ones((batch_size, input_length), dtype=torch.long)
    logger.info("Batch size = {:d}".format(batch_size))
    logger.info("The length of input tokens = {:d}".format(input_length))

    with torch.no_grad():
        for i in range(total_iters):
            if i == warmup_iters:
                start = time.perf_counter()
            model(example_inputs)
            cur_accelerator.synchronize()
        end = time.perf_counter()
    latency = (end - start) / ((total_iters - warmup_iters) * batch_size)
    throughput = ((total_iters - warmup_iters) * batch_size) / (end - start)
    logger.info("Latency: {:.3f} ms".format(latency * 10**3))
    logger.info("Throughput: {:.3f} samples/sec".format(throughput))



# Copyright (c) 2024 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from neural_compressor.torch.algorithms import Quantizer
from neural_compressor.torch.algorithms.mixed_low_precision.modules import HPUMixedPrecisionLinear
from neural_compressor.torch.algorithms.weight_only.modules import HPUWeightOnlyLinear


class HybridGPTQQuantizer(Quantizer):
    def __init__(self, quant_config):
        super().__init__(quant_config)
        if isinstance(quant_config, dict):
            json_file = [cfg.json_file for cfg in quant_config.values()]
            assert len(json_file) > 0, "Cannot get json file from config."
            self.quant_config = json_file[0]

    def prepare(self, model):
        _prepare(model)
        return model

    def convert(self, model):
        _convert(model)
        return model


def set_module(model, op_name, new_module):
    """Set module with a given op name.

    Args:
        model (object): the input model.
        op_name (str): name of op.
        new_module (object): the input model.

    Returns:
        module (object).
    """
    module = model
    name_list = op_name.split(".")
    for name in name_list[:-1]:
        if hasattr(module, name):
            module = getattr(module, name)
        else:
            module = module
    setattr(module, name_list[-1], new_module)


def _convert(model):
    for name, module in model.named_modules():
        # replace `HPUWeightOnlyLinear`s forward func
        if isinstance(module, HPUWeightOnlyLinear):
            module = HPUMixedPrecisionLinear.convert_from_weight_only(module)
            set_module(model, name, module)

    return model


def _prepare(model):
    for name, module in model.named_modules():
        # replace `HPUWeightOnlyLinear`s forward func
        if isinstance(module, HPUWeightOnlyLinear):
            module = HPUMixedPrecisionLinear.prepare_from_weight_only(module)
            set_module(model, name, module)

    return model



# Copyright (c) 2024 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import torch
import types
from .._quant_common.quant_config import ScaleFormat
from .common import is_runtime_scale_patching
from neural_compressor.torch.utils.auto_accelerator import auto_detect_accelerator

cur_device = auto_detect_accelerator().current_device_name()


def add_scale_registry(patched_mod):
    """Update scale registry"""
    patched_mod.scale_members = set()
    patched_mod.register_scale = types.MethodType(register_scale, patched_mod)
    return patched_mod


def register_scale(patched_mod, name, scale, scale_format):
    """Register the scale name into patched_mod.scale_member_list so that the scalar scale is updated into state_dict"""
    if name in patched_mod.scale_members:
        raise ValueError("scale member {} already exists".format(name))
    scale = create_scale_tensor(scale, scale_format)
    patched_mod.scale_members.add(name)
    setattr(patched_mod, name, scale)

def create_scale_tensor(orig_scales, scale_format):
    if scale_format not in ScaleFormat.__members__.values():
        raise ValueError(f"Invalid scale format: {scale_format}")

    # dynamic quantization case, need to avoid it if possible - see SW-230996
    if orig_scales is None:
        return
    try:
        scale_creation_func = scale_to_cpu if is_runtime_scale_patching() else _scale_creation_funcs_map[scale_format]
    except KeyError:
        raise KeyError(f"Scale format {scale_format} isn't in _scale_creation_funcs_map")

    if isinstance(orig_scales, (torch.Tensor, float)):
        return scale_creation_func(orig_scales)
    elif isinstance(orig_scales, list):
        return [scale_creation_func(x) for x in orig_scales]
    else:
        raise ValueError("unexpected scale format value {}".format(scale_format))

def scale_to_cpu(scale_tensor):
    # Note: If the tensor has only one element, create a torch scalar (0-dim tensor).
    # Otherwise, it cause a hang in the MoE op. See SW-239190 for details.
    if scale_tensor.numel() == 1:
        scale_tensor = torch.tensor(scale_tensor.item(), dtype=scale_tensor.dtype)
    return scale_tensor.to("cpu")

def scale_to_const(scale_tensor):
    return torch.nn.Parameter(scale_tensor, requires_grad=False)

# scalar scale is a performance optimization for LLM layers in small BS
def scale_to_scalar(scale):
    if isinstance(scale, torch.Tensor):  # tensor case
        if scale.numel() == 1:
            return scale.item()
        else:
            raise Exception("scale as scalar isn't supported for scale tensors of dim > 0")
    elif isinstance(scale, float):  # already scalar case
        return scale
    else:
        raise Exception(f"Unexpected scale instance type: {type(scale).__name__}, expected Torch.tensor or float number")

_scale_creation_funcs_map = {ScaleFormat.SCALAR: scale_to_scalar, ScaleFormat.CONST: scale_to_const}

def get_scale_dtype(scale):
    if isinstance(scale, torch.Tensor):  # tensor case
        return scale.dtype
    elif isinstance(scale, float):  # already scalar case
        return type(scale).__name__
    elif scale is None: # possible dynamic scalar case
        return None
    else:
        raise Exception(f"Unexpected scale instance type: {type(scale).__name__}, expected Torch.tensor or float number")


def get_param_scales_from_scalar(patched_mod, prefix, dtype=torch.bfloat16, device=cur_device):
    """Get all scales in param_list, used for saving scalar scales"""
    scale_dict = {}
    for name in patched_mod.scale_members:
        if hasattr(patched_mod, name) and isinstance(getattr(patched_mod, name), float):
            scale_dict.update({
                # E.g. lm_head.scale_input
                prefix + name: torch.tensor(getattr(patched_mod, name), dtype=dtype, device=device),
            })
    return scale_dict


def get_param_scales_from_list(patched_mod, prefix, dtype=torch.bfloat16, device=cur_device):
    """Get all scales in param_list, used for saving scalar scales"""
    scale_dict = {}
    for name in patched_mod.scale_members:
        if hasattr(patched_mod, name) and isinstance(getattr(patched_mod, name), list):
            scale_dict.update({
                # E.g. lm_head.scale_input
                prefix + name: torch.cat(
                    [torch.tensor(v, dtype=dtype, device=device) for v in getattr(patched_mod, name)]
                ),
            })
    return scale_dict


def set_param_scales_as_scalar(patched_mod, state_dict):
    """Set all scales in param_list, used for loading scalar scales"""
    state_dict_keys = list(state_dict.keys())
    for name in patched_mod.scale_members:
        if hasattr(patched_mod, name) and isinstance(getattr(patched_mod, name), float):
            for k in state_dict_keys:
                if name == k.split('.')[-1]:
                    v = state_dict.pop(k)
                    setattr(patched_mod, name, v.item())
    return state_dict


def set_param_scales_into_list(patched_mod, state_dict):
    """Set all scales in param_list, used for loading scalar scales"""
    state_dict_keys = list(state_dict.keys())
    for name in patched_mod.scale_members:
        if hasattr(patched_mod, name) and isinstance(getattr(patched_mod, name), list):
            for k in state_dict_keys:
                if name == k.split('.')[-1]:
                    v = state_dict.pop(k)
                    if patched_mod.scale_format == ScaleFormat.SCALAR:
                        assert v.dim() == 2, "Unexpected scale shape, please raise an issue."
                        v = [v[i].item() for i in range(v.size(0))]
                    else:
                        v = [v[i] for i in range(v.size(0))]
                    setattr(patched_mod, name, v)
    return state_dict


def get_state_dict(patched_mod, *args, destination=None, prefix='', keep_vars=False):
    """replace torch.nn.Module.state_dict"""
    cur_state_dict = torch.nn.Module.state_dict(patched_mod, *args, destination=destination, prefix=prefix, keep_vars=keep_vars)
    device = cur_device
    dtype = patched_mod.hp_dtype
    if patched_mod.scale_format == ScaleFormat.SCALAR:
        scale_dict = get_param_scales_from_scalar(patched_mod, prefix, dtype=dtype, device=device)
        cur_state_dict.update(scale_dict)
    get_param_scales_from_list(patched_mod, prefix, dtype=dtype, device=device)
    return cur_state_dict


def load_state_dict(patched_mod, state_dict, prefix, local_metadata, strict,
                    missing_keys, unexpected_keys, error_msgs):
    """replace torch.nn.Module._load_from_state_dict"""
    if patched_mod.scale_format == ScaleFormat.SCALAR:
        state_dict = set_param_scales_as_scalar(patched_mod, state_dict)
    set_param_scales_into_list(patched_mod, state_dict)
    torch.nn.Module._load_from_state_dict(patched_mod, state_dict, prefix, local_metadata, strict,
                    missing_keys, unexpected_keys, error_msgs)


def update_state_dict_method(patched_mod):
    """update fetch and load state_dict method for scalar scales"""
    patched_mod.state_dict = types.MethodType(get_state_dict, patched_mod)
    patched_mod._load_from_state_dict = types.MethodType(load_state_dict, patched_mod)



"""Shared utilities for FP8 quantization configuration and scale files."""

# Copyright (c) 2025 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import functools
import json
import os

import numpy as np
import torch
from enum import Enum, auto
from functools import lru_cache 
from ..utils.logger import logger
from neural_compressor.torch.algorithms.fp8_quant.model_configs import ModuleConfig
from neural_compressor.torch.utils.auto_accelerator import auto_detect_accelerator

cur_device = auto_detect_accelerator().current_device_name()

UNMEASURED_MODELS = "UnmeasuredModels"

def dequant_original_fp8_weight_if_needed(mod: torch.nn.Module, param: torch.Tensor) -> torch.Tensor:
    """Dequantize FP8 weights using the module hook when required.

    Args:
        mod (torch.nn.Module): Module that may provide a dequantization callback.
        param (torch.Tensor): Parameter tensor to inspect for FP8 data types.

    Returns:
        torch.Tensor: The original parameter or a dequantized replacement.
    """
    if param.dtype in [torch.float8_e4m3fn]:
        if hasattr(mod, "get_dequant_weights_func"):
            dequant_weights_func = mod.get_dequant_weights_func()
            if dequant_weights_func is not None:
                param = dequant_weights_func(mod)
            else:
                raise RuntimeError(f"Got fp8 weight for {mod}, but dequant function is None, please check.")
        else:
            RuntimeError(f"Got fp8 weight for {mod}, but dequant function is not found, please check.")
            
    return param

class QuantTensorType(Enum):
    """Enum describing the type of quantized tensor representation.

    Attributes:
        MEASUREMENTS: Tensor values captured from calibration measurements.
        CONST: Constant quantized tensors derived from stored parameters.
        DYNAMIC: Dynamically computed quantized tensors at runtime.
    """

    MEASUREMENTS = auto()
    CONST = auto()
    DYNAMIC = auto()


class ShapeList:
    """Container for transporting shape information through format conversions.

    Attributes:
        data (list[int] | None): Shape dimensions carried with the wrapper.
    """

    data = None


def rec_fn(x, fn):
    """Recursively apply a function to nested container values.

    Args:
        x (Any): Input object that can be a dict, list, tuple, or leaf value.
        fn (Callable[[Any], Any]): Function to apply to each leaf value.

    Returns:
        Any: Structure with the same container layout and transformed leaves.
    """
    if isinstance(x, dict):
        return {k: rec_fn(x[k], fn) for k in x}
    elif isinstance(x, list):
        return [rec_fn(k, fn) for k in x]
    elif isinstance(x, tuple):
        return tuple([rec_fn(k, fn) for k in x])
    else:
        return fn(x)


def save_json(d, fname):
    """Save a Python dictionary to a JSON file.

    Args:
        d (dict): Data to serialize.
        fname (str): Destination file path.
    """
    with open(fname, "w") as f:
        json.dump(d, f, indent=4)


def load_json(fname):
    """Load a JSON file into a Python dictionary.

    Args:
        fname (str): Source file path.

    Returns:
        dict: Deserialized JSON content.
    """
    with open(fname, "r") as f:
        d = json.load(f)
    return d


def save_npz(d, fname):
    """Save a Python object to a NumPy NPZ archive.

    Args:
        d (Any): Object to store in the archive.
        fname (str): Destination file path.
    """
    np.savez(fname, d)


def load_npz(fname):
    """Load a Python object from a NumPy NPZ archive.

    Warning:
        NumPy may deserialize pickled objects during load. Only use trusted files.

    Args:
        fname (str): Source file path to the NPZ archive.

    Returns:
        Any: The stored object from the archive.
    """
    d = np.load(fname, allow_pickle=True)
    return d["arr_0"].item()


def save_file(model, d, source_format, fname, mode, num_samples=0):
    """Persist a scale-related payload to disk in the chosen format.

    Args:
        model (torch.nn.Module): Model associated with the data.
        d (dict): Source data keyed by module name.
        source_format (type): In-memory data format type.
        fname (str): Target file path.
        mode (str): Label describing the saved data (e.g., "Scale").
        num_samples (int, optional): Optional number of calibration samples. Defaults to 0.
    """
    from .._quant_common.quant_config import get_hqt_config
    config = get_hqt_config(model)
    logger.debug("Saving %s file: %s", mode, fname)
    ext = os.path.splitext(fname)[1]
    target_format = file_functions[ext]['format']
    dc = rec_fn(d, format_functions[(source_format, target_format)])
    df = {
        "GlobalRank": config.cfg["global_rank"],
        "LocalRank": config.cfg["local_rank"],
        "Mode": mode,
        "Nodes": dc,
    }
    if num_samples > 0:
        df = { "NumSamples": num_samples, **df}
    try:
        file_functions[ext]['save'](df, fname)
    except:
        pass


def load_file(fname, target_format, fail_on_file_not_exist):
    """Load a scale file and convert it to module configuration objects.

    Args:
        fname (str): Source file path.
        target_format (type): Desired data format type for loaded content.
        fail_on_file_not_exist (bool): Whether to raise if the file is missing.

    Returns:
        dict[str, ModuleConfig]: Mapping of module names to configuration objects.
    """
    logger.debug("Loading file: %s", fname)
    ext = os.path.splitext(fname)[1]
    source_format = file_functions[ext]['format']
    d = {}
    if os.path.isfile(fname):
        d = file_functions[ext]['load'](fname)
    elif fail_on_file_not_exist:
        raise FileNotFoundError(f"Failed to load file {fname}")
    if "Nodes" in d:
        dc = {k: ModuleConfig(**fix_fields(d["Nodes"][k])) for k in d["Nodes"]}
        dc = {k: module_convert(dc[k], format_functions[(source_format, target_format)]) for k in dc}
    else:
        dc = {}
    return dc


# convert module config data to other format
def module_convert(m, fcn):
    """Convert a module configuration to the specified data format.

    Args:
        m (ModuleConfig): Module configuration to convert.
        fcn (Callable[[Any], Any]): Conversion function applied to all tensor-like values.

    Returns:
        ModuleConfig: Converted configuration instance.
    """
    mt = ModuleConfig(
        tuple([fcn(x) for x in m.inputs]),
        (
            tuple(
                [fcn(m.outputs)],
            )
            if type(m.outputs) == np.ndarray
            else tuple([fcn(y) for y in m.outputs])
        ),
        {k: fcn(m.params[k]) for k in m.params},
    )
    return mt


def fix_fields(d):
    """Normalize legacy field names in serialized configuration data.

    Args:
        d (dict): Input dictionary to update in-place.

    Returns:
        dict: The updated dictionary with normalized keys.
    """
    if "input" in d:
        d["inputs"] = d.pop("input")
    if "output" in d:
        d["outputs"] = d.pop("output")
    return d


def save_scales(model, d, source_format, fname):
    """Saves scales measured of a given model.

    Args:
        model : The measured model.
        d : Modules_names to configuration dictionary.
        source_format : How the data is stored in memory.
        fname : File to save the scales to.
    """
    dc = {k: d[k].__dict__ for k in d}
    save_file(model, dc, source_format, fname, "Scale")


def load_scales(fname, target_format):
    """Loads scales from given file.

    Args:
        fname : File to load the scales from.
        target_format: How the data is stored in file.
    """
    logger.debug("Loading scales file %s", fname)
    d = load_file(fname, target_format, False)
    return d


def convert_scales_to_tensors_dict(scales_obj, scales_file_format, hp_dtype, device=cur_device):
    """Convert stored scale objects to tensors on the requested device.

    Args:
        scales_obj (dict[str, ModuleConfig]): Module scales loaded from disk.
        scales_file_format (type): Format used in the stored file.
        hp_dtype (torch.dtype): Target tensor dtype.
        device (str, optional): Device identifier to move tensors to. Defaults to current device.

    Returns:
        dict[str, ModuleConfig]: Scales represented as tensors in the target dtype/device.
    """
    scales_temp = {k: scales_obj[k].__dict__ for k in scales_obj}
    scales_temp = format_functions_rec((scales_file_format, torch.Tensor))(scales_temp)
    scales_temp = rec_fn(scales_temp, lambda x: x.to(dtype=hp_dtype, device=device))
    scales = {k: ModuleConfig(**scales_temp[k]) for k in scales_temp}
    return scales


file_functions = {
    ".json": {'format': list, 'save': save_json, 'load': load_json},
    ".npz": {'format': np.ndarray, 'save': save_npz, 'load': load_npz}
}

format_functions = {
    (torch.Tensor, torch.Tensor): lambda x: x,
    (np.ndarray, np.ndarray): lambda x: x,
    (list, list): lambda x: x,
    (torch.Tensor, np.ndarray): lambda x: x.detach().cpu().float().numpy(),
    (torch.Tensor, list): lambda x: x.detach().cpu().float().numpy().tolist(),
    (np.ndarray, torch.Tensor): torch.tensor,
    (np.ndarray, list): lambda x: x.tolist(),
    (list, torch.Tensor): torch.tensor,
    (list, np.ndarray): lambda x: np.array(x),
    (list, ShapeList): lambda x: [int(s) for s in x[0]],
}


format_functions_rec = lambda k: functools.partial(rec_fn, fn=format_functions[k])


def get_device_type_for_scales(mod):
    """Return the device name used for storing scales for a module.

    Args:
        mod (torch.nn.Module): Module whose quantization config is consulted.

    Returns:
        str: Device type string for scale storage.
    """
    from .._quant_common.quant_config import get_hqt_config
    config = get_hqt_config(mod).cfg
    return config["device_for_scales"]


@lru_cache
def is_runtime_scale_patching():
    """Check whether runtime scale patching is enabled via environment variable.

    Returns:
        bool: True when runtime patching is enabled.
    """
    return os.getenv("RUNTIME_SCALE_PATCHING", "False").lower() in ["true", "1"]

#TODO [SW-224612]: Use cguid to calc scales and remove the check
@lru_cache
def is_calc_scale_with_cguid():
    """Check whether scale calculation uses cguid logic.

    Returns:
        bool: True when cguid-based scale calculation is enabled.
    """
    return os.getenv("CALC_SCALE_WITH_CGUID", "True").lower() in ["true", "1"]

#TODO [SW-224612]: Use cguid to calc scales and remove the check
@lru_cache
def is_calc_scale_rounding_with_cguid():
    """Check whether scale rounding uses cguid-based configuration.

    Returns:
        bool: True when rounding with cguid is enabled.
    """
    return is_calc_scale_with_cguid() and os.getenv("CALC_ROUNDING_WITH_CGUID", "False").lower() in ["true", "1"]



# Copyright (c) 2024 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import torch
from enum import Enum
from .common import ModuleConfig
from neural_compressor.torch.utils.auto_accelerator import auto_detect_accelerator, INCAcceleratorType
from neural_compressor.torch.utils import logger

cur_accelerator = auto_detect_accelerator()

descale_fcn = lambda x, scale: torch.mul(x, scale)
scale_fcn = lambda x, scale: torch.div(x, scale)
cast_fcn = lambda x, dtype: x.to(dtype=dtype)
cast_to_fp8_fcn = lambda x, dtype, scale_inv=None: torch.ops.hpu.cast_to_fp8_v2(x, scale_inv, False, False, dtype)[0]
def calculate_scale_maxabs_with_cguid(x, maxMode, **kwargs):
    return torch.ops.hpu.calculate_scale_for_cast(
        x, maxMode.value, ScaleCalculationRoundingMode.NO_SCALE_ROUNDING.value, **kwargs
    )


def calculate_scale_rounding_with_cguid(x, scaleMode, **kwargs):
    return torch.ops.hpu.calculate_scale_for_cast(
        x, ScaleCalculationMaxMode.NO_MAX_CALCULATION.value, scaleMode.value, **kwargs
    )


class ScaleCalculationMaxMode(Enum):
    NO_MAX_CALCULATION = 0
    MAX_ABS_PTS_CALCULATION = 1
    MAX_ABS_PCS_CALCULATION = 2


class ScaleCalculationRoundingMode(Enum):
    NO_SCALE_ROUNDING = 0
    SCALE_TO_POW2_ROUNDING = 1

GAUDI2 = INCAcceleratorType.GAUDI2
GAUDI3 = INCAcceleratorType.GAUDI3

EXP_WIDTH = {
    torch.float32: 8,
    torch.bfloat16: 8,
    torch.float8_e4m3fn: 4,
    torch.float8_e5m2: 5,
}

descale_fcn = lambda x, scale: torch.mul(x, scale)
scale_fcn = lambda x, scale: torch.div(x, scale)
cast_fcn = lambda x, dtype: x.to(dtype=dtype)
cast_to_fp8_fcn = lambda x, dtype, scale_inv=None: torch.ops.hpu.cast_to_fp8_v2(x, scale_inv, False, False, dtype)[0]
cast_from_fp8_fcn = lambda x, dtype, scale=None: torch.ops.hpu.cast_from_fp8(x, scale, dtype)
quantize_per_tensor_to_fp8 = lambda x, scale, zero_point, quant_min, quant_max, dtype=None, axis=None: torch.ops.quantized_decomposed.quantize_per_tensor(
    x, scale, zero_point, quant_min, quant_max, dtype=dtype
)
dequantize_per_tensor_from_fp8 = lambda x, scale, zero_point, quant_min, quant_max, dtype, out_dtype=None, axis=None: torch.ops.quantized_decomposed.dequantize_per_tensor(
    x, scale, zero_point, quant_min, quant_max, dtype=dtype, out_dtype=out_dtype
)
quantize_per_channel_to_fp8 = lambda x, scale, zero_point, axis, quant_min, quant_max, dtype=None: torch.ops.quantized_decomposed.quantize_per_channel(
    x, scale, zero_point, axis, quant_min, quant_max, dtype=dtype
)
dequantize_per_channel_from_fp8 = lambda x, scale, zero_point, axis, quant_min, quant_max, dtype, out_dtype=None: torch.ops.quantized_decomposed.dequantize_per_channel(
    x, scale, zero_point, axis, quant_min, quant_max, dtype=dtype, out_dtype=out_dtype
)


def get_default_exp_bias(dtype):
    exp_width = EXP_WIDTH[dtype]
    return 2 ** (exp_width - 1) - 1


EXP_BIAS_SETS = {
    (GAUDI2, torch.float8_e4m3fn): [3, 7, 11, 15],
    (GAUDI2, torch.float8_e5m2): [15],
    (GAUDI3, torch.float8_e4m3fn): range(0, 63),
    (GAUDI3, torch.float8_e5m2): range(0, 63),
}

MAX_RANGE = {
    torch.float32: torch.finfo(torch.float32).max,
    torch.bfloat16: torch.finfo(torch.bfloat16).max,
    torch.float8_e4m3fn: torch.finfo(torch.float8_e4m3fn).max,
    # float8_e4m3fn data type is 8-bit floating point consist of Exponent: 4, Mantissa: 3, bias: 7. It's supported by Gaudi3.
    torch.float8_e5m2: torch.finfo(torch.float8_e5m2).max
    # float8_e5m2 data type is 8-bit floating point consist of Exponent: 5, Mantissa: 2, bias: 15. IEEE 754, with NaN and inf.
}
# TODO FSW-12066 solve fp_utils
try:
    MAX_RANGE[torch.float8_e4m3fnuz] = torch.finfo(torch.float8_e4m3fnuz).max
    # float8_e4m3fnuz data type is 8-bit floating point consist of Exponent: 4, Mantissa: 3, bias: 8 with 1 sign bit. It's supported by Gaudi2.
except AttributeError as e:
    pass

def get_fullscale(dtype, device, exp_bias=None):
    default_exp_bias = get_default_exp_bias(dtype)
    fullscale = 1
    if device == GAUDI2 and dtype == torch.float8_e4m3fn:
        # TODO FSW-12066 solve fp_utils
        try:
            fullscale = MAX_RANGE[torch.float8_e4m3fnuz]
        except AttributeError as e:
            pass
    else:
        fullscale = MAX_RANGE[dtype]
    exp_bias = default_exp_bias if exp_bias is None else exp_bias
    fullscale = fullscale * (2 ** (default_exp_bias - exp_bias))
    return float(fullscale)


def get_fullscales_by_expbias_set(dtype, device, expbias_set):
    return [get_fullscale(dtype, device, exp_bias=eb) for eb in expbias_set]


def get_fp8_hw_alligned_scales_by_device(dtype, device):
    if device not in [GAUDI2, GAUDI3]:
        logger.warning("hw aligned scales not supported for device {}".format(device))
        return None # only Gaudis support hw aligned scales
    exp_bias_set = EXP_BIAS_SETS.get((device, dtype), None)
    return (
        None
        if exp_bias_set is None
        else [x / get_fullscale(dtype, device) for x in get_fullscales_by_expbias_set(dtype, device, exp_bias_set)]
    )

def get_fp8_hw_alligned_scales(dtype):
    inc_device_type = auto_detect_accelerator().get_inc_accelerator_type()
    return get_fp8_hw_alligned_scales_by_device(dtype, inc_device_type)

DEVICES_SCALE_FACTORS = {
    INCAcceleratorType.GAUDI2: 4,
    INCAcceleratorType.GAUDI3: 1,
}
FP8_143_SCALES = {
    device: get_fp8_hw_alligned_scales_by_device(torch.float8_e4m3fn, device) for device in DEVICES_SCALE_FACTORS.keys()
}
FP8_143_SCALES_TRAITS = {
    device: (
        min(FP8_143_SCALES[device]),
        max(FP8_143_SCALES[device]),
        DEVICES_SCALE_FACTORS[device],
    )
    for device in DEVICES_SCALE_FACTORS.keys()
}

def calc_scale_from_maxabs(xmaxabs, fullscale, backoff=1):
    scale = xmaxabs / (fullscale * backoff)
    return scale

def mmse_scale_multi(x, ref_scale, scales, lp_dtype, hp_dtype):
    if not scales:
        raise ValueError(
            "got empty scale list. it is possible that scale method isn't supported by current device."
        )
    # TODO: SW-176672 move weights to hpu before the scale calculations
    x = x.to("hpu")
    Nch = x.shape[-1]
    opt_err = torch.ones(Nch, dtype=hp_dtype, device=x.device) * torch.inf
    opt_scale = torch.ones(Nch, dtype=hp_dtype, device=x.device) * -1
    sum_axis = list(range(x.ndim - 1))
    rs = ref_scale.unsqueeze(dim=1).transpose(0, 1)
    for s in scales:
        sv = torch.ones(Nch, dtype=hp_dtype, device=x.device) * s
        xscales = rs * sv
        y = scale_fcn(x, xscales)
        y = cast_to_fp8_fcn(y, lp_dtype)
        cur_accelerator.synchronize()  # we are measuring the error so we want to avoid fusion of the converts
        y = cast_fcn(y, hp_dtype)
        y = descale_fcn(y, xscales)
        err = torch.sum((x - y) ** 2, dim=sum_axis)
        opt_scale = torch.where(err < opt_err, sv, opt_scale)
        opt_err = torch.where(err < opt_err, err, opt_err)
        cur_accelerator.synchronize()
    return opt_scale * ref_scale


def mmse_scale(x, scales, lp_dtype, hp_dtype):
    if not scales:
        raise ValueError(
            "got empty scale list. it is possible that scale method isn't supported by current device."
        )
    # TODO: SW-176672 move weights to hpu before the scale calculations
    x = x.to("hpu")
    opt_err = torch.ones(1, dtype=hp_dtype, device=x.device) * torch.inf
    opt_scale = -1
    for s in scales:
        y = scale_fcn(x, s)
        y = cast_to_fp8_fcn(y, lp_dtype)
        cur_accelerator.synchronize()  # we are measuring the error so we want to avoid fusion of the converts
        y = cast_fcn(y, hp_dtype)
        y = descale_fcn(y, s)
        err = torch.norm(x - y)
        opt_scale = torch.where(err <= opt_err, s, opt_scale)
        opt_err = torch.where(err <= opt_err, err, opt_err)
        cur_accelerator.synchronize()
    return opt_scale


def manipulate_scales(scales, func):
    """Applies a function to the inputs, outputs, and weights of the ModuleConfig object."""
    new_inputs = [func(input) for input in scales.inputs]
    new_outputs = [func(output) for output in scales.outputs]
    new_weights = {}
    if "weight" in scales.params.keys():
        if isinstance(scales.params["weight"], (torch.Tensor, float)):
            new_weights = {"weight": func(scales.params["weight"])}
        elif isinstance(scales.params["weight"], dict):
            new_weights_dict = {}
            for key, wt in scales.params["weight"].items():
                new_weights_dict[key] = func(wt)
            new_weights = {"weight": new_weights_dict}
    new_scales = ModuleConfig(new_inputs, new_outputs, new_weights)
    return new_scales


def invert_scale(x):
    """Inverts the scale of the input tensor, list of tensors, or tuple of tensors."""
    def invert(x):
        if isinstance(x, torch.Tensor):
            return torch.reciprocal(x)
        return 1.0 / x
    if x is None:
        return None
    if isinstance(x, (list, tuple)):
        return type(x)(invert(x_i) for x_i in x)
    return invert(x)


def invert_scales(scales):
    """Inverts the scales of the input ModuleConfig object."""
    return manipulate_scales(scales, invert_scale)



# Copyright (c) 2025 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from enum import Enum, auto

from .round_scales_function import *
from ..common import get_device_type_for_scales
from .scales_method import *
from ...utils.logger import logger
from .scale_method_config import ScaleRoundMethod, ScaleValueType, ScaleGranularity, CfgStr

class QuantTensorName(Enum):
    INPUT = auto()
    WEIGHT_OUT_CH = auto()
    WEIGHT_IN_CH = auto()
    OUTPUT = auto()


HW_ALIGNED_ROUND_METHODS = [ScaleRoundMethod.HW_ALIGNED, ScaleRoundMethod.HW_ALIGNED_FIXED]

class ScaleMethodFactory:
    def __init__(self, config, params, mod, op_type):
        # config is dict mapping weight and activation to their ScaleMethodConfig(scale.py) which defines ScaleMethodString
        # params is `get_hqt_config(model).cfg["scale_params"]`, we copy in order not to damage it for other modules
        self.params = params.copy()
        self.mod = mod
        self.device_for_scales = get_device_type_for_scales(mod)
        self.op_type = op_type

        input_scale_method_config = config[CfgStr.ACTIVATION]
        weight_scale_method_config = config[CfgStr.WEIGHT]
        output_scale_method_config = input_scale_method_config

        self.scale_method_config_map = {
            QuantTensorName.INPUT: input_scale_method_config,
            QuantTensorName.WEIGHT_OUT_CH: weight_scale_method_config,
            QuantTensorName.WEIGHT_IN_CH: weight_scale_method_config,
            QuantTensorName.OUTPUT: output_scale_method_config,
        }

        logger.trace("%s %s", self.__class__.__name__, self.__dict__)

    def get_scale_method(self, tensor_type, is_dynamic=False):
        backoff = 1.0 if is_dynamic else self.scale_method_config_map[tensor_type].backoff
        scale_round_method = self.scale_method_config_map[tensor_type].rounding_method
        scale_value_type = self.scale_method_config_map[tensor_type].scale_value_type
        scale_granularity = self.scale_method_config_map[tensor_type].granularity

        # create scale_round_method according to ScaleRoundMethod enum name
        scale_round_method_name = scale_round_method.name
        if scale_round_method in HW_ALIGNED_ROUND_METHODS:
            # HW_ALIGNED rounding methods use device for scales
            scale_round_method = scale_round_method_registry[scale_round_method_name](self.device_for_scales)
        else:
            scale_round_method = scale_round_method_registry[scale_round_method_name]()

        logger.trace(
            "get_scale_method backoff=%s scale_round_method=%s scale_value_type=%s scale_granularity=%s op_type=%s is_dynamic=%s",
            backoff,
            scale_round_method,
            scale_value_type,
            scale_granularity,
            self.op_type,
            is_dynamic,
        )

        match (scale_value_type, scale_granularity, tensor_type, self.op_type):
            ## dummy
            case (ScaleValueType.DUMMY_SCALES, _, _, _) if tensor_type not in [QuantTensorName.WEIGHT_IN_CH]:
                return DummyScales(scale_round_method, self.params, self.device_for_scales)
            ## fixed value
            case (ScaleValueType.FIXED_VALUE, _, _, _) if tensor_type not in [QuantTensorName.WEIGHT_IN_CH]:
                return FixedScale(scale_round_method, self.params, self.device_for_scales)
            ## output max abs and opt, for linear and matmul
            case (_, _, QuantTensorName.OUTPUT, _) \
                if self.op_type in {"linear", "matmul"}:
                    if scale_value_type in {ScaleValueType.MAXABS, ScaleValueType.OPT}:
                        return MulAdditionalScales(scale_round_method, self.params, self.device_for_scales)
            ## maxabs/opt in channel PTS
            case (_, ScaleGranularity.PTS, QuantTensorName.WEIGHT_IN_CH, _):
                return None
            case (ScaleValueType.MAXABS, ScaleGranularity.PTS, _, _):
                if is_dynamic:
                    return MaxAbsDynamicPts(scale_round_method, self.params, self.device_for_scales, backoff)
                return MaxAbsPts(scale_round_method, self.params, self.device_for_scales, backoff)
            ## maxabs/opt in channel PCS
            case (_, ScaleGranularity.PCS, QuantTensorName.WEIGHT_IN_CH, _)\
                if scale_value_type in {ScaleValueType.MAXABS, ScaleValueType.OPT}:
                in_channel_size = self.mod.weight.shape[1]
                return InputChannelScale(scale_round_method, self.params, self.device_for_scales, in_channel_size)
            ## maxabs PCS
            case (ScaleValueType.MAXABS, ScaleGranularity.PCS, _, _):
                if is_dynamic:
                    return MaxAbsDynamicPcs(scale_round_method, self.params, self.device_for_scales, backoff)
                return MaxAbsPcs(scale_round_method, self.params, self.device_for_scales, backoff)
            ## opt PTS
            case (ScaleValueType.OPT, ScaleGranularity.PTS, _, _):
                opt_list_of_scales = self.scale_method_config_map[tensor_type].params["weight_scales"]
                return OptScalesPts(scale_round_method, opt_list_of_scales, self.params, self.device_for_scales, backoff)
            case (ScaleValueType.OPT, ScaleGranularity.PCS, _, _):
                opt_list_of_scales = self.scale_method_config_map[tensor_type].params["weight_scales"]
                return OptScalesPcs(scale_round_method, opt_list_of_scales, self.params, self.device_for_scales, backoff)
            case _:
                raise NotImplementedError("the config: scale_round_method: " + \
                                          str(scale_round_method) +
                                          ", scale_value_type: " + str(scale_value_type) +
                                          ", scale_granularity: " + str(scale_granularity) +
                                          " in tensor: " + str(tensor_type) +
                                          " op type "  + self.op_type + " not implemented ")


# Copyright (c) 2024 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Save and load the quantized model."""


# pylint:disable=import-error
import torch

try:
    import intel_extension_for_pytorch as ipex
except:  # pragma: no cover
    assert False, "Please install IPEX for smooth quantization."

from neural_compressor.torch.algorithms.static_quant import load, save


def recover_model_from_json(model, json_file_path, example_inputs):  # pragma: no cover
    """Recover ipex model from JSON file.

    Args:
        model (object): fp32 model need to do quantization.
        json_file_path (json): configuration JSON file for ipex.
        example_inputs (tuple or torch.Tensor or dict): example inputs that will be passed to the ipex function.

    Returns:
        model (object): quantized model
    """
    from torch.ao.quantization.observer import MinMaxObserver

    if ipex.__version__ >= "2.1.100":
        qconfig = ipex.quantization.get_smooth_quant_qconfig_mapping(alpha=0.5, act_observer=MinMaxObserver)
    else:
        qconfig = ipex.quantization.get_smooth_quant_qconfig_mapping(alpha=0.5, act_observer=MinMaxObserver())
    if isinstance(example_inputs, dict):
        model = ipex.quantization.prepare(model, qconfig, example_kwarg_inputs=example_inputs, inplace=True)
    else:
        model = ipex.quantization.prepare(model, qconfig, example_inputs=example_inputs, inplace=True)

    model.load_qconf_summary(qconf_summary=json_file_path)
    model = ipex.quantization.convert(model, inplace=True)
    with torch.no_grad():
        try:
            if isinstance(example_inputs, dict):
                # pylint: disable=E1120,E1123
                model = torch.jit.trace(model, example_kwarg_inputs=example_inputs)
            else:
                model = torch.jit.trace(model, example_inputs)
            model = torch.jit.freeze(model.eval())
        except:
            if isinstance(example_inputs, dict):
                # pylint: disable=E1120,E1123
                model = torch.jit.trace(model, example_kwarg_inputs=example_inputs, strict=False, check_trace=False)
            else:
                model = torch.jit.trace(model, example_inputs, strict=False)
            model = torch.jit.freeze(model.eval())
        if isinstance(example_inputs, dict):
            model(**example_inputs)
            model(**example_inputs)
        elif isinstance(example_inputs, tuple) or isinstance(example_inputs, list):
            model(*example_inputs)
            model(*example_inputs)
        else:
            model(example_inputs)
            model(example_inputs)
    return model



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
#
# Copyright (c) 2025 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Utils for quantization."""

import types
from typing import Any

import torch
import torch.nn as nn

from .quant_linear import QuantLinear
from .tensor_quantizer import TensorQuantizer


def convert(module: nn.Module, quant_cfg=None, quant_module=None):
    """Convert the model to a quantized one with quant config."""

    # update class
    original_cls = type(module)
    module.__class__ = quant_module
    module.forward = types.MethodType(quant_module.forward, module)

    # setup quantizers
    module._setup(quant_cfg)

    return module


def replace_with_quant_linear(model, quant_cfg=None):
    """Recursively replace the module with quantized module."""

    # TODO: support more modules, like kv.
    for name, child in model.named_children():
        if isinstance(child, nn.Linear):
            if "lm_head" in name:
                continue
            # REPLACE on the parent (model), not on child
            quantized = convert(child, quant_cfg, QuantLinear)
            setattr(model, name, quantized)

        # now recurse into whichever module is now at `model.name`
        replace_with_quant_linear(getattr(model, name), quant_cfg=quant_cfg)

    return model


def get_quant_config_with_scheme(scheme: str):
    """Get quantization config."""

    try:
        # use scheme definitions from AutoRound since we utilize the quantization functions now
        from auto_round.schemes import preset_name_to_scheme

        quant_cfg = preset_name_to_scheme(scheme)
        return quant_cfg
    except ImportError:
        return None


def convert_model_with_mapping(model, mapping=None):
    """Process mapping to quant config."""
    # key is torch module, TODO: support more key format, like layer name.
    for key in mapping:
        # TODO: support more torch modules
        if key == nn.Linear:
            quant_cfg = get_quant_config_with_scheme(mapping[key])
            if quant_cfg is None:
                continue
            replace_with_quant_linear(model, quant_cfg)

    replaced_modules = sum(isinstance(m, TensorQuantizer) for _, m in model.named_modules())
    print(f"Inserted {replaced_modules} quantizers")


def get_quant_config(scheme: str) -> dict[str, Any]:
    """Generate quantization config for a torch model.

    Args:
        model: The PyTorch model to analyze

    Returns:
        Dictionary containing the quantization configuration
    """

    # TODO: support more quant config
    try:
        from auto_round.export.export_to_llmcompressor.config import initialize_quantization

        quantization_config = initialize_quantization(scheme=scheme)
        quantization_config = quantization_config.to_dict()
        quantization_config["provider"] = "auto-round"
        quantization_config["config_groups"]["group_0"]["weights"]["is_mx"] = True
        quantization_config["config_groups"]["group_0"]["input_activations"]["is_mx"] = True
        quantization_config["format"] = "float-quantized"

    except ImportError:
        quantization_config = None

    return quantization_config


def get_quantization_format(module) -> str | None:
    """Gets the quantization string.

    Gets the quantization string by iterating through the module and its children.
    The first non-None quantization string is returned.
    """

    def _get_quantization_from_layer(layer):
        weight_quantizer = getattr(layer, "weight_quantizer", None)
        input_quantizer = getattr(layer, "input_quantizer", None)

        if weight_quantizer is None or weight_quantizer._disabled:
            return None

        # TODO: support more quant format
        if weight_quantizer.num_bits == 8 and weight_quantizer.data_type == "mx_fp8":
            return "MXFP8"

        if weight_quantizer.num_bits == 4 and weight_quantizer.data_type == "mx_fp4":
            return "MXFP4"

        # Raise error for unsupported num_bits
        raise NotImplementedError(f"Unsupported quantizer with num_bits: {weight_quantizer.num_bits}")

    quantization = _get_quantization_from_layer(module)
    if quantization is not None:
        return quantization

    for _, layer in module.named_children():
        format = get_quantization_format(layer)
        if format is not None:
            return format

    return None


def is_quantlinear(module: nn.Module) -> bool:
    """Returns whether the module is a quantized linear layer."""
    return "QuantLinear" in type(module).__name__



# Copyright (c) 2024 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Intel Neural Compressor PyTorch utilities."""

import enum
import importlib
import json
import os
from collections import UserDict
from typing import Callable, Dict, List, Optional, Tuple, Union

import psutil
import torch
import torch.nn as nn
from typing_extensions import TypeAlias

from neural_compressor.common.utils import (
    Mode,
    ProcessorType,
    Statistics,
    cpu_info,
    detect_processor_type_based_on_hw,
    logger,
)
from neural_compressor.torch.utils import SaveLoadFormat, is_optimum_habana_available, is_transformers_imported

if is_transformers_imported():
    import transformers

    SUPPORTED_LAYERS = [nn.Linear, transformers.pytorch_utils.Conv1D]
else:
    SUPPORTED_LAYERS = [nn.Conv1d, nn.Linear]

OP_NAME_AND_TYPE_TUPLE_TYPE: TypeAlias = Tuple[str, Union[torch.nn.Module, Callable]]

# Dictionary to store a mapping between algorithm names and corresponding algo implementation(function)
algos_mapping: Dict[str, Callable] = {}


# All constants for torch
WHITE_MODULE_LIST = [torch.nn.Linear, torch.nn.Conv1d, torch.nn.Conv2d, torch.nn.Conv3d]
UNIT_MAPPING = {"KB": 1024, "MB": 1024**2, "GB": 1024**3, "TB": 1024**4, "B": 1}

HPU_SAFE_WEIGHTS_NAME = "hpu_model.safetensors"
WEIGHT_NAME = "quantized_weight.pt"
HPU_WEIGHT_NAME = "quantized_hpu_weight.pt"
SHARDED_WEIGHT_NAME = "model_bin_index.json"
SHARDED_HPU_WEIGHT_NAME = "quantized_hpu_weight.safetensors"
QCONFIG_NAME = "qconfig.json"


def register_algo(name):
    """Decorator function to register algorithms in the algos_mapping dictionary.

    Usage example:
        @register_algo(name=example_algo)
        def example_algo(model: torch.nn.Module, quant_config: RTNConfig) -> torch.nn.Module:
            ...

    Args:
        name (str): The name under which the algorithm function will be registered.

    Returns:
        decorator: The decorator function to be used with algorithm functions.
    """

    def decorator(algo_func):
        algos_mapping[name] = algo_func
        return algo_func

    return decorator


def fetch_module(model, op_name):
    """Get module with a given op name.

    Args:
        model (object): the input model.
        op_name (str): name of op.

    Returns:
        module (object).
    """
    module = model
    name_list = op_name.split(".")
    for name in name_list:
        if hasattr(module, name):
            module = getattr(module, name)
        else:
            logger.warning(f"The {op_name} is not present in the model.")
            return None
    return module


def set_module(model, op_name, new_module):
    """Set module with a given op name.

    Args:
        model (object): the input model.
        op_name (str): name of op.
        new_module (object): the input model.

    Returns:
        module (object).
    """
    name_list = op_name.split(".")
    if len(name_list) == 1:
        setattr(model, name_list[-1], new_module)
        return
    else:
        second_last_module = fetch_module(model, ".".join(name_list[:-1]))
        if second_last_module is None:
            logger.warning(f"Setting skipped as the {op_name} is not present in the model.")
            return None
        else:
            setattr(second_last_module, name_list[-1], new_module)


get_attr = fetch_module
set_attr = set_module


def get_model_info(model: torch.nn.Module, white_module_list: List[Callable]) -> List[Tuple[str, str]]:
    """Get model info according to white_module_list."""
    module_dict = dict(model.named_modules())
    filter_result = []
    filter_result_set = set()
    for op_name, module in module_dict.items():
        if isinstance(module, tuple(white_module_list)):
            pair = (op_name, type(module).__name__)
            if pair not in filter_result_set:
                filter_result_set.add(pair)
                filter_result.append(pair)
    logger.debug(f"Get model info: {filter_result}")
    return filter_result


def get_double_quant_config_dict(double_quant_type="BNB_NF4"):
    """Query config dict of double_quant according to double_quant_type.

    Args:
        double_quant_type (str, optional): double_quant type. Defaults to "BNB_NF4".
    """
    from neural_compressor.torch.utils.constants import DOUBLE_QUANT_CONFIGS

    assert double_quant_type in DOUBLE_QUANT_CONFIGS, "Supported double quant configs: {}".format(
        list(DOUBLE_QUANT_CONFIGS.keys())
    )
    return DOUBLE_QUANT_CONFIGS[double_quant_type]


def get_quantizer(model, quantizer_cls, quant_config=None, *args, **kwargs):
    """Get the quantizer.

    Initialize a quantizer or get `quantizer` attribute from model.

    Args:
        model (torch.nn.Module): pytorch model.
        quantizer_cls (Quantizer): quantizer class of a specific algorithm.
        quant_config (dict, optional): Specifies how to apply the algorithm on the given model.
            Defaults to None.

    Returns:
        quantizer object.
    """
    if not hasattr(model, "quantizer"):
        quantizer = quantizer_cls(quant_config=quant_config, *args, **kwargs)
        return quantizer
    else:
        return model.quantizer


def postprocess_model(model, mode, quantizer):
    """Process `quantizer` attribute of model according to current phase.

    In `prepare` phase, the `quantizer` is set as an attribute of the model
    to avoid redundant initialization during `convert` phase.

    In 'convert' or 'quantize' phase, the unused `quantizer` attribute is removed.

    Args:
        model (torch.nn.Module): pytorch model.
        mode (Mode): The mode of current phase, including 'prepare', 'convert' and 'quantize'.
        quantizer (Quantizer): quantizer object.
    """
    if mode == Mode.PREPARE:
        model.quantizer = quantizer
    elif mode == Mode.CONVERT or mode == Mode.QUANTIZE:
        if getattr(model, "quantizer", False):
            del model.quantizer


def dump_model_op_stats(mode, tune_cfg):
    """Dump quantizable ops stats of model to user.

    Args:
        mode (object): quantization mode.
        tune_cfg (dict): quantization config
    """
    if mode == Mode.PREPARE:
        return
    res = {}
    # collect all dtype info and build empty results with existing op_type
    dtype_set = set()
    for op, config in tune_cfg.items():
        op_type = op[1]
        config = config.to_dict()
        if not config["dtype"] == "fp32":
            num_bits = config["bits"]
            group_size = config["group_size"]
            dtype_str = "A32W{}G{}".format(num_bits, group_size)
            dtype_set.add(dtype_str)
    dtype_set.add("FP32")
    dtype_list = list(dtype_set)
    dtype_list.sort()

    for op, config in tune_cfg.items():
        config = config.to_dict()
        op_type = op[1]
        if op_type not in res.keys():
            res[op_type] = {dtype: 0 for dtype in dtype_list}

    # fill in results with op_type and dtype
    for op, config in tune_cfg.items():
        config = config.to_dict()
        if config["dtype"] == "fp32":
            res[op_type]["FP32"] += 1
        else:
            num_bits = config["bits"]
            group_size = config["group_size"]
            dtype_str = "A32W{}G{}".format(num_bits, group_size)
            res[op_type][dtype_str] += 1

    # update stats format for dump.
    field_names = ["Op Type", "Total"]
    field_names.extend(dtype_list)
    output_data = []
    for op_type in res.keys():
        field_results = [op_type, sum(res[op_type].values())]
        field_results.extend([res[op_type][dtype] for dtype in dtype_list])
        output_data.append(field_results)

    Statistics(output_data, header="Mixed Precision Statistics", field_names=field_names).print_stat()


def get_model_device(model: torch.nn.Module):
    """Get the device.

    Args:
        model (torch.nn.Module): the input model.

    Returns:
        device (str): a string.
    """
    for n, p in model.named_parameters():
        return p.data.device.type  # p.data.device == device(type='cpu')


def get_processor_type_from_user_config(user_processor_type: Optional[Union[str, ProcessorType]] = None):
    """Get the processor type.

    Get the processor type based on the user configuration or automatically detect it based on the hardware.

    Args:
        user_processor_type (Optional[Union[str, ProcessorType]]): The user-specified processor type. Defaults to None.

    Returns:
        ProcessorType: The detected or user-specified processor type.

    Raises:
        AssertionError: If the user-specified processor type is not supported.
        NotImplementedError: If the processor type is not recognized.
    """
    if user_processor_type is None:
        processor_type = detect_processor_type_based_on_hw()
    elif isinstance(user_processor_type, ProcessorType):
        processor_type = user_processor_type
    elif isinstance(user_processor_type, str):
        user_processor_type = user_processor_type.lower().capitalize()
        assert user_processor_type in ProcessorType.__members__, f"Unsupported processor type: {user_processor_type}"
        processor_type = ProcessorType(user_processor_type)
    else:
        raise NotImplementedError(f"Unsupported processor type: {user_processor_type}")
    return processor_type


def dowload_hf_model(repo_id, cache_dir=None, repo_type=None, revision=None):
    """Download hugging face model from hf hub."""
    from huggingface_hub.constants import DEFAULT_REVISION, HUGGINGFACE_HUB_CACHE
    from huggingface_hub.file_download import REGEX_COMMIT_HASH, repo_folder_name
    from huggingface_hub.utils import EntryNotFoundError

    if cache_dir is None:
        cache_dir = HUGGINGFACE_HUB_CACHE
    if revision is None:
        revision = DEFAULT_REVISION
    if repo_type is None:
        repo_type = "model"
    storage_folder = os.path.join(cache_dir, repo_folder_name(repo_id=repo_id, repo_type=repo_type))
    commit_hash = None
    if REGEX_COMMIT_HASH.match(revision):
        commit_hash = revision
    else:
        ref_path = os.path.join(storage_folder, "refs", revision)
        if os.path.exists(ref_path):
            with open(ref_path) as f:
                commit_hash = f.read()
    if storage_folder and commit_hash:
        pointer_path = os.path.join(storage_folder, "snapshots", commit_hash)
        if os.path.isdir(pointer_path) and any(
            file.endswith(".bin") or file.endswith(".safetensors") for file in os.listdir(pointer_path)
        ):
            return pointer_path
    from huggingface_hub import list_repo_files, snapshot_download

    files_info = list_repo_files(repo_id)
    ignore_patterns = (
        ["*.bin", "*.bin.index.json"]
        if (
            any(file for file in files_info if file.endswith(".bin"))
            and any(file for file in files_info if file.endswith(".safetensors"))
        )
        else None
    )

    file_path = snapshot_download(repo_id, ignore_patterns=ignore_patterns)
    return file_path


def load_empty_model(pretrained_model_name_or_path, cls=None, **kwargs):
    """Load a empty model."""
    from accelerate import init_empty_weights
    from transformers import AutoConfig, AutoModelForCausalLM
    from transformers.models.auto.auto_factory import _BaseAutoModelClass

    cls = AutoModelForCausalLM if cls is None else cls
    is_local = os.path.isdir(pretrained_model_name_or_path)
    if is_local:  # pragma: no cover
        path = pretrained_model_name_or_path
    else:
        path = dowload_hf_model(pretrained_model_name_or_path)
    if cls.__base__ == _BaseAutoModelClass:
        with init_empty_weights():
            model = cls.from_pretrained(path, **kwargs)
    else:  # pragma: no cover
        config = cls.config_class.from_pretrained(path, **kwargs)
        with init_empty_weights():
            model = cls(config, **kwargs)
    model.tie_weights()
    model.eval()
    model.path = pretrained_model_name_or_path
    return model


def get_module(module, key):
    """Get module from model by key name.

    Args:
        module (torch.nn.Module): original model
        key (str): module name to be replaced
    """
    name_list = key.split(".")
    for name in name_list:
        module = getattr(module, name, None)
    return module


def get_layer_names_in_block(model, supported_types=SUPPORTED_LAYERS, to_quant_block_names=None):
    """Retrieves the names of layers within each block of the model.

    Returns:
        list: A list of strings, where each string is the name of a layer
              within a block of the model.
    """
    for n, m in model.named_modules():
        if isinstance(m, tuple(supported_types)):
            m.tmp_name = n
    layers_in_block = []
    if bool(to_quant_block_names):
        all_blocks = to_quant_block_names
    else:
        all_blocks = get_block_names(model)
    for block_names in all_blocks:
        for block_name in block_names:
            block = get_module(model, block_name)
            for n, m in block.named_modules():
                if hasattr(m, "tmp_name"):
                    layers_in_block.append(m.tmp_name)
    for n, m in model.named_modules():
        if hasattr(m, "tmp_name"):
            delattr(m, "tmp_name")
    return layers_in_block


def to_dtype(input, dtype=torch.float32):  # pragma: no cover
    """Moves input data to the specified data type.

    Args:
    input: The input data to be moved.
    dtype: The target data type.

    Returns:
    The input data on the specified data type.
    """
    if input is None:
        return None
    if isinstance(input, torch.Tensor):
        return input.to(dtype)
    if isinstance(input, dict) or isinstance(input, UserDict):
        for inp in input.keys():
            input[inp] = to_dtype(input[inp], dtype)

    elif isinstance(input, list) or isinstance(input, tuple):
        if len(input) == 0:
            return input
        input_res = []
        for inp in input:
            input_res.append(to_dtype(inp, dtype))
        if isinstance(input, tuple):
            input_res = tuple(input_res)
        input = input_res

    return input


# for VLM usage
def to_device(input, device=torch.device("cpu")):  # pragma: no cover
    """Moves input data to the specified device.

    Args:
    input: The input data to be moved.
    device: The target device.

    Returns:
    The input data on the specified device.
    """
    if input is None:
        return None
    if isinstance(input, torch.Tensor):
        return input.to(device)
    if isinstance(input, dict) or isinstance(input, UserDict):
        for inp in input.keys():
            input[inp] = to_device(input[inp], device)
    elif isinstance(input, list) or isinstance(input, tuple):
        if len(input) == 0:
            return input
        input_res = []
        for inp in input:
            input_res.append(to_device(inp, device))
        if isinstance(input, tuple):
            input_res = tuple(input_res)
        input = input_res

    return input


def get_block_names(model):
    """Get the block names for transformers-like networks.

    Args:
    model: The model.

    Returns:
    block_names: A list whose elements are list of block's layer names
    """
    block_names = []
    target_modules = []
    for n, m in model.named_modules():
        if hasattr(type(m), "__name__") and "ModuleList" in type(m).__name__:
            target_modules.append((n, m))
            break  ## only find the first modulelist, may be not robust
    for i, target_m in enumerate(target_modules):
        block_names.append([])
        for n, m in target_m[1].named_children():
            block_names[i].append(target_m[0] + "." + n)
    return block_names


def validate_modules(module_names):  # pragma: no cover
    """Test a list of modules' validity.

    Args:
    modules (list of str): List of strings to be validated.

    Returns:
    bool: True if all modules have equal length or not dependent, otherwise False.
    """
    if not bool(module_names):
        raise ValueError("Empty modules")
    if len(module_names) < 2:
        return True
    split_modules = [s.split(".") for s, _ in module_names]
    lengths = [len(parts) for parts in split_modules]
    if len(set(lengths)) == 1:
        return True
    max_length = max(lengths)
    min_length = min(lengths)
    longest_module = next(s for s in split_modules if len(s) == max_length)
    shortest_module = next(s for s in split_modules if len(s) == min_length)
    shortest_module = ".".join(shortest_module)
    longest_module = ".".join(longest_module)
    # Check if the shortest name is a substring of the longest name
    if shortest_module in longest_module:
        raise ValueError(
            "Invalid modules, at least two modules detected" " as dependent, {shortest_module} and {longest_module}"
        )
    return True


def get_multimodal_block_names(model, quant_vision=False):
    """Get the multimodal model block names for transformers-like networks.

    Args:
    model: The model.

    Returns:
    block_names: A list whose elements are list of block's layer names
    """
    block_names = []
    target_modules = []
    Vison_blocks_tuple = (
        "vision",
        "visual",
    )
    for n, m in model.named_modules():
        if hasattr(type(m), "__name__") and "ModuleList" in type(m).__name__:
            if quant_vision or all(key not in n.lower() for key in (Vison_blocks_tuple)):
                target_modules.append((n, m))
    validate_modules(target_modules)
    for i, target_m in enumerate(target_modules):
        block_names.append([])
        for n, m in target_m[1].named_children():
            block_names[i].append(target_m[0] + "." + n)
    return block_names


def detect_device(device=None):  # pragma: no cover
    """Detects the device to use for model execution (GPU, HPU, or CPU).

    Args:
        device (str, int, torch.device, optional):
            - If a string ('cuda', 'cpu', or 'hpu') or torch.device is provided, that device is selected.
            - If an integer is provided, it treats it as a GPU device index.
            - If None or 'auto', it automatically selects 'cuda' if available, 'hpu' if Habana is available,
              or falls back to 'cpu'.

    Returns:
        str: The selected device in string format ('cuda:X', 'hpu', or 'cpu').
    """

    def is_valid_digit(s):
        try:
            num = int(s)
            return 0 <= num
        except:
            return False

    dev_idx = None
    if is_valid_digit(device):
        dev_idx = int(device)
        device = "auto"
    if device is None or device == "auto":
        if torch.cuda.is_available():
            device = torch.device("cuda")
            print("Using GPU device")
        elif is_optimum_habana_available():
            device = torch.device("hpu")
            print("Using HPU device")
        # Use CPU as a fallback
        else:
            device = torch.device("cpu")
            print("Using CPU device")
        if dev_idx is not None and str(device) != "cpu":
            device = str(device) + f":{dev_idx}"
        return str(device)
    elif isinstance(device, torch.device):
        device = str(device)
    return device


def find_matching_blocks(model, all_blocks, to_quant_block_names=None):
    """Find and return matching blocks in the model based on to_quant_block_names.

    Args:
        model: The model (not used in this specific function but kept for completeness).
        all_blocks: List of lists, where each inner list contains full block names in the model.
        to_quant_block_names: Comma-separated string of target block names to match.

    Returns:
        target_blocks: List of lists containing full paths of matching blocks in the model.
    """
    import re

    if not to_quant_block_names:
        return all_blocks
    to_quant_block_list = to_quant_block_names
    if isinstance(to_quant_block_names, list) or isinstance(to_quant_block_names, tuple):
        return to_quant_block_names
    if isinstance(to_quant_block_names, str):
        to_quant_block_list = [name.strip() for name in to_quant_block_names.split(",")]
    target_blocks = []
    for block_list in all_blocks:
        matched_sublist = []
        for name in to_quant_block_list:
            matches = [block for block in block_list if re.search(name, block)]
            if matches:
                matched_sublist.extend(matches)
        if matched_sublist:
            target_blocks.append(matched_sublist)
        if not target_blocks:
            raise ValueError(
                "No block names matched. Please check the input for to_quant_block_name,"
                "or set to_quant_block_name to None to automatically match quantizable blocks."
            )
    return target_blocks


def get_non_persistent_buffers(model):
    """Get all non-persistent buffers in the model.

    Args:
        model (torch.nn.Module): PyTorch model

    Returns:
        dict: A dictionary containing all non-persistent buffers, {buffer_names: buffer_tensors}
    """
    non_persistent_buffers = {}
    for name, module in model.named_modules():
        # Check the module's non-persistent buffer
        for buffer_name, buffer in module._buffers.items():
            if buffer_name in module._non_persistent_buffers_set:
                non_persistent_buffers[(name, buffer_name)] = buffer
    return non_persistent_buffers


def load_non_persistent_buffers(model, non_persistent_buffers):
    """Load all non-persistent buffers into the model.

    Args:
        model (torch.nn.Module): PyTorch model
        non_persistent_buffers (dict): A dictionary containing all non-persistent buffers, {buffer_names: buffer_tensors}
    """
    for full_name, buffer in non_persistent_buffers.items():
        module_name, buffer_name = full_name
        module = model.get_submodule(module_name) if module_name else model
        setattr(module, buffer_name, buffer)


# copied from neural_compressor/adaptor/torch_utils/util.py
def move_input_device(input, device="cpu"):
    """Auto mapping input to device for all kinds of format.

    Args:
        input (torch.tensor): input data
        device (str, optional): target device. Defaults to "cpu".

    Returns:
        input (torch.tensor): input data on target device
    """
    if isinstance(input, dict) or isinstance(input, UserDict):
        tmp_input = {}
        for k, inp in input.items():
            tmp_input[k] = move_input_device(inp, device)
        input = tmp_input
    elif isinstance(input, list) or isinstance(input, tuple):
        tmp_input = []
        for inp in input:
            tmp_input.append(move_input_device(inp, device))
        input = tmp_input
    elif isinstance(input, torch.Tensor):
        input = input.to(device)  # pylint: disable=no-member
    return input


# copied from neural_compressor/adaptor/torch_utils/util.py
def forward_wrapper(model, input):
    """Model forward with device auto mapping.

    Args:
        model (torch.nn.Module): input model
        input (torch.tensor): input data

    Returns:
        output: output data
    """
    try:
        device = next(model.parameters()).device
    except:
        # for RecursiveScriptModule
        device = "cpu"
    input = move_input_device(input, device)
    if isinstance(input, dict) or isinstance(input, UserDict):
        output = model(**input)
    elif isinstance(input, list) or isinstance(input, tuple):
        try:
            output = model(*input)
        except:
            output = model(input)
    else:
        output = model(input)
    return output


def get_enum_from_format(format):
    """Make sure Save&Load format is an Enum object."""
    if isinstance(format, SaveLoadFormat):
        return format
    for obj in SaveLoadFormat:
        if format == obj.value:
            return obj
        elif format.upper() == obj.name:
            return obj
    raise ValueError(f"Invalid format value ('{format}'). Enter one of [{[m.name for m in SaveLoadFormat]}]")


def read_json_file(file_path):
    """Read a JSON file and return its content."""
    if not file_path or not os.path.exists(file_path):
        raise FileNotFoundError(f"File {file_path} does not exist.")
    with open(file_path, "r", encoding="utf-8") as f:
        return json.load(f)


def write_json_file(file_path, data):
    """Write data to a JSON file."""
    if not file_path:
        raise ValueError("File path cannot be empty.")
    with open(file_path, "w", encoding="utf-8") as f:
        json.dump(data, f, indent=4, ensure_ascii=False)
    logger.info(f"Data written to {file_path} successfully.")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2023 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Utils for layer wise quantization."""

import gc
import json
import os

import torch
from accelerate.utils import set_module_tensor_to_device
from safetensors import safe_open
from safetensors.torch import save_file

from neural_compressor.common import options
from neural_compressor.torch.algorithms.weight_only.modules import INCWeightOnlyLinear
from neural_compressor.torch.utils import is_hpu_available
from neural_compressor.torch.utils.utility import dowload_hf_model

from .load import load

LWQ_WORKSPACE = os.path.join(options.workspace, "lwq_tmpdir")


class QDQLayer(torch.nn.Module):
    """Quantized and Dequantized Layer."""

    def __init__(self, module, input_scale=None) -> None:
        """Init the QDQLayer object."""
        super().__init__()
        self.quant = torch.ao.quantization.QuantStub()
        self.module = module
        self.dequant = torch.ao.quantization.DeQuantStub()
        self.input_scale = input_scale

    def forward(self, X):
        """Forward function."""
        if self.input_scale is not None:
            X = torch.mul(X, self.input_scale)
        X = self.quant(X)
        X = self.module(X)
        X = self.dequant(X)
        return X


def get_module(model, key):
    """Get module from model by key name.

    Args:
        model (torch.nn.Module): original model
        key (str): module name to be replaced
    """
    attrs = key.split(".")
    module = model
    for attr in attrs:
        try:
            attr = int(attr)
            module = module[attr]
        except:
            module = getattr(module, attr)
    return module


def get_children(model):
    """Get all the children of given model."""
    module_list = []
    children = list(model.children())
    if len(children) == 0:
        return [model]
    for child in children:
        module_list += get_children(child)
    return module_list


def get_named_children(model, pre=[]):
    """Get all the name and children of given model."""
    module_list = []
    if len(list(model.children())) == 0:
        return [(".".join(pre), model)]
    for name, module in model.named_children():
        module_list += get_named_children(module, pre=pre + [name])
    return module_list


def get_super_module_by_name(model, module_name):
    """Get the father module with given name of child module."""
    name_list = module_name.split(".")
    for name in name_list[:-1]:
        if hasattr(model, name):
            model = getattr(model, name)
        else:  # pragma: no cover
            return None
    if hasattr(model, name_list[-1]):
        return model
    else:  # pragma: no cover
        return None


def update_module(model, module_name, new_module):
    """Update module."""
    super_module = get_super_module_by_name(model, module_name)
    if super_module:
        setattr(super_module, module_name.split(".")[-1], new_module)


def load_layer_wise_quantized_model(path):  # pragma: no cover
    """Load layer wise quantized model."""
    model = torch.load(os.path.join(path, "model_arch.pt"))
    for name, _ in model.named_modules():
        if name + ".pt" in os.listdir(path):
            update_module(model, name, torch.load(os.path.join(path, name + ".pt")))
    model.eval()
    return model


def load_tensor_from_shard(pretrained_model_name_or_path, tensor_name, prefix=None):  # pragma: no cover
    """Load tensor from shard."""
    path = _get_path(pretrained_model_name_or_path)
    idx_dict = json.load(open(os.path.join(path, "pytorch_model.bin.index.json"), "r"))["weight_map"]
    if tensor_name not in idx_dict.keys():
        if tensor_name.replace(f"{prefix}.", "") in idx_dict.keys():
            tensor_name = tensor_name.replace(f"{prefix}.", "")
        else:
            assert False, "{} not in the index.json".format(tensor_name)
    return load_tensor(os.path.join(path, idx_dict[tensor_name]), tensor_name, None)


def load_tensor(path, tensor_name=None, prefix=None):
    """Load a tensor from bin file with given tensor name."""
    # transformers.modeling_utils
    if tensor_name:
        if "gamma" in tensor_name:  # pragma: no cover
            tensor_name = tensor_name.replace("gamma", "weight")
        if "beta" in tensor_name:  # pragma: no cover
            tensor_name = tensor_name.replace("beta", "bias")

    if os.path.isdir(path):
        path = os.path.join(path, "pytorch_model.bin")
    state_dict = load(path, tensor_name, prefix)
    if tensor_name:
        if tensor_name in state_dict:
            return state_dict[tensor_name]
        else:  # pragma: no cover
            return state_dict[tensor_name.replace(f"{prefix}.", "")]
    else:  # pragma: no cover
        return state_dict


def load_tensor_from_safetensors(path, tensor_name=None, prefix=None, device="cpu"):
    """Load a tensor from safetensors file with given tensor name."""
    with safe_open(path, framework="pt", device=device) as f:
        if tensor_name in f.keys():
            value = f.get_tensor(tensor_name)
        elif prefix and tensor_name.replace(f"{prefix}.", "") in f.keys():
            value = f.get_tensor(tensor_name.replace(f"{prefix}.", ""))
        else:
            raise ValueError(f"Tensor '{tensor_name}' not found in the file '{path}'")
    return value


def load_tensor_from_safetensors_shard(
    pretrained_model_name_or_path, tensor_name, prefix=None, device="cpu"
):  # pragma: no cover
    """Load tensor from shard."""
    path = _get_path(pretrained_model_name_or_path)
    idx_dict = json.load(open(os.path.join(path, "model.safetensors.index.json"), "r"))["weight_map"]
    if tensor_name not in idx_dict.keys():
        if tensor_name.replace(f"{prefix}.", "") in idx_dict.keys():
            tensor_name = tensor_name.replace(f"{prefix}.", "")
        else:
            assert False, "{} not in the index.json".format(tensor_name)
    return load_tensor_from_safetensors(os.path.join(path, idx_dict[tensor_name]), tensor_name, device)


def _get_path(pretrained_model_name_or_path):
    is_local = os.path.isdir(pretrained_model_name_or_path)
    if is_local:  # pragma: no cover
        path = pretrained_model_name_or_path
    else:
        path = dowload_hf_model(pretrained_model_name_or_path)
    return path


get_path = _get_path


def load_value(model, param_name, path, device="cpu"):
    """Load the module value.

    Args:
        model (torch.nn.module): torch model.
        param_name (str): module name.
        path (str): path to load state_dict per layer.
        device (str, optional): module device. Defaults to "cpu".

    Returns:
        tensor: the module value.
    """
    if "lm_head" in param_name and getattr(model.config, "tie_word_embeddings", True):
        input_embeddings = model.get_input_embeddings()
        modules = get_named_children(model)
        for name, module in modules:
            if module == input_embeddings:
                param_name = name + "." + param_name.split(".")[-1]
    prefix = model.base_model_prefix
    files = os.listdir(path)
    safetensors_files = [filename for filename in files if filename.endswith(".safetensors")]

    if device == torch.device("hpu"):
        device = "hpu"

    if len(safetensors_files) == 1:
        value = load_tensor_from_safetensors(
            os.path.join(path, "model.safetensors"), param_name, prefix=prefix, device=device
        )
    elif len(safetensors_files) >= 2:
        value = load_tensor_from_safetensors_shard(path, param_name, prefix=prefix, device=device)
    elif "pytorch_model.bin.index.json" in files:
        value = load_tensor_from_shard(path, param_name, prefix)
    else:
        value = load_tensor(os.path.join(path, "pytorch_model.bin"), param_name, prefix)
    return value


def load_module(model, module_name, path, device="cpu"):
    """Load all named parameters of module.

    Args:
        model (torch.nn.module): torch model.
        module_name (str): module name.
        path (str): path to load state_dict per layer.
        device (str, optional): module device. Defaults to "cpu".
    """
    module = get_module(model, module_name)
    for n, p in module.named_parameters():
        param_name = module_name + "." + n
        value = load_value(model, param_name, path, device)
        set_module_tensor_to_device(model, param_name, device, value)


def load_first_layer_only(user_model, model_name):
    """Load first layer only.

    Args:
        user_model (torch.nn.Module): input model
        model_name (str): model name or path
    """
    for name, m in user_model.named_modules():
        if ("layers" not in name or "layers.0" in name) and len(name) > 0 and len(list(m.named_children())) == 0:
            load_module(user_model, name, get_path(model_name), device="hpu" if is_hpu_available() else "cpu")


def register_weight_hooks(model, path, device="cpu", clean_weight=True, saved_path=None, indicated_layers=None):
    """Register weight hooks for model.

    Args:
        model (torch.nn.module): torch model.
        path (str): path to load state_dict per layer.
        device (str, optional): module device. Defaults to "cpu".
        clean_weight (bool, optional): to clean model weight. Defaults to True.
        saved_path (str, optional): path to save module weight. Defaults to None.

    Returns:
        list: handlers.
    """
    if saved_path:
        os.makedirs(saved_path, exist_ok=True)

    def forward_pre_hook(name):
        def hook(module, input):
            state_dict = None
            if os.path.exists(os.path.join(LWQ_WORKSPACE, f"{name}.pt")):
                state_dict = torch.load(
                    os.path.join(LWQ_WORKSPACE, f"{name}.pt"),
                    map_location=torch.device(device) if isinstance(device, str) else device,
                )
            for n, p in module.named_parameters():
                param_name = name + "." + n
                if state_dict:
                    value = state_dict[n]
                else:
                    value = load_value(model, param_name, path, device=device)
                set_module_tensor_to_device(model, param_name, device, value)
            module = module.to(device)

        return hook

    def forward_hook(name):
        def hook(module, input, output):
            if saved_path:
                file_path = os.path.join(saved_path, f"{name}.pt")
                torch.save(module.state_dict(), file_path)
            clean_module_weight(module)

        return hook

    handle = {}
    modules = get_named_children(model)
    for name, module in modules:
        if indicated_layers is not None and name not in indicated_layers:  # pragma: no cover
            # load other layers to memory
            state_dict = None
            if os.path.exists(os.path.join(LWQ_WORKSPACE, f"{name}.pt")):
                state_dict = torch.load(
                    os.path.join(LWQ_WORKSPACE, f"{name}.pt"),
                    map_location=torch.device(device) if isinstance(device, str) else device,
                )
            for n, p in module.named_parameters():
                param_name = name + "." + n
                if state_dict:
                    value = state_dict[n]
                else:
                    value = load_value(model, param_name, path, device=device)
                set_module_tensor_to_device(model, param_name, device, value)
            module = module.to(device)
            continue
        handle[name] = [module.register_forward_pre_hook(forward_pre_hook(name))]
        if clean_weight:
            handle[name] += [module.register_forward_hook(forward_hook(name))]
    return handle


def clean_module_weight(module):
    """Clean module weight."""
    if isinstance(module, QDQLayer):
        submodule = module.module
    else:
        submodule = module

    if isinstance(module, INCWeightOnlyLinear):
        for n, m in submodule._buffers.items():
            old_value = getattr(submodule, n)
            with torch.no_grad():
                submodule._buffers[n] = torch.zeros(old_value.shape, device="meta")

    for n, m in submodule.named_parameters():
        is_buffer = n in submodule._buffers
        old_value = getattr(submodule, n)
        with torch.no_grad():
            if is_buffer:
                submodule._buffers[n] = torch.zeros(old_value.shape, device="meta")
            else:
                param_cls = type(submodule._parameters[n])
                kwargs = submodule._parameters[n].__dict__
                if is_hpu_available():
                    from habana_frameworks.torch.core import weight_sharing

                    if param_cls == weight_sharing.HabanaParameterWrapper:
                        try:
                            kwargs.pop("change_device_placement")
                        except KeyError:
                            pass

                new_value = torch.zeros(old_value.shape, device="meta")
                new_value = param_cls(new_value, requires_grad=old_value.requires_grad, **kwargs).to("meta")
                submodule._parameters[n] = new_value
    # gc.collect()


def save_layers_in_shards_iteratively(checkpoint_dir, output_dir, layers_per_shard=10):
    """Save model layers iteratively in shards, each shard containing a fixed number of layers using safetensors."""
    os.makedirs(output_dir, exist_ok=True)

    # Get list of checkpoint files in the checkpoint_dir
    checkpoint_files = [f for f in os.listdir(checkpoint_dir) if f.endswith(".pt")]
    checkpoint_files.sort()

    bin_index = {}
    shard_idx = 0
    layer_buffer = []

    # Iterate over each checkpoint file and process layer state dicts
    for checkpoint_file in checkpoint_files:
        layer_path = os.path.join(checkpoint_dir, checkpoint_file)
        print(f"Loading layer from {layer_path}")

        # Load the layer checkpoint
        checkpoint = torch.load(layer_path, map_location="cpu")
        layer_state_dict = checkpoint

        # Add the layer's state dict to the buffer
        layer_buffer.append(layer_state_dict)

        # If the buffer has reached the number of layers per shard, save the shard
        if len(layer_buffer) == layers_per_shard:
            shard_state_dict = {}
            for i, layer_dict in enumerate(layer_buffer):
                shard_state_dict.update(layer_dict)
                # Update the bin index for each layer
                for layer_name in layer_dict.keys():
                    bin_index[layer_name] = shard_idx

            # Save the shard to disk using safetensors
            shard_filename = f"model_shard-{str(shard_idx + 1).zfill(5)}-of-{str((len(checkpoint_files) // layers_per_shard) + 1).zfill(5)}.safetensors"
            shard_path = os.path.join(output_dir, shard_filename)
            save_file(shard_state_dict, shard_path)  # Save using safetensors
            print(f"Saved shard {shard_idx + 1} of {len(checkpoint_files) // layers_per_shard + 1} at {shard_path}")

            # Clear the buffer and move to the next shard
            layer_buffer.clear()
            shard_idx += 1

    # If there are any remaining layers in the buffer, save them as the last shard
    if layer_buffer:
        shard_state_dict = {}
        for i, layer_dict in enumerate(layer_buffer):
            shard_state_dict.update(layer_dict)
            # Update the bin index for each layer
            for layer_name in layer_dict.keys():
                bin_index[layer_name] = shard_idx

        # Save the final shard
        shard_filename = f"model_shard-{str(shard_idx + 1).zfill(5)}-of-{str((len(checkpoint_files) // layers_per_shard) + 1).zfill(5)}.safetensors"
        shard_path = os.path.join(output_dir, shard_filename)
        save_file(shard_state_dict, shard_path)  # Save using safetensors
        print(f"Saved final shard {shard_idx + 1} of {len(checkpoint_files) // layers_per_shard + 1} at {shard_path}")

    # Save bin index to a JSON file
    bin_index_file = os.path.join(output_dir, "model_bin_index.json")
    with open(bin_index_file, "w") as f:
        json.dump(bin_index, f, indent=4)

    print(f"Model bin index saved to {bin_index_file}")


from safetensors.torch import load_file  # Safetensors load function


def load_model_from_shards_with_safetensors(shard_dir, bin_index_file):
    """Load the model from its shards and the bin index using safetensors.

    Args:
        shard_dir (str): Directory containing the model shard files.
        bin_index_file (str): Path to the bin index JSON file.

    Returns:
        torch.nn.Module: The reconstructed model with the layers.
    """
    # Load bin index to get the layer -> shard mapping
    with open(bin_index_file, "r") as f:
        bin_index = json.load(f)

    full_state_dict = {}

    # Sort and load the shard files
    shard_files = [f for f in os.listdir(shard_dir) if f.endswith(".safetensors")]
    shard_files.sort()

    for shard_file in shard_files:
        shard_path = os.path.join(shard_dir, shard_file)
        print(f"Loading shard from {shard_path}")
        shard_state_dict = load_file(shard_path, device="hpu" if is_hpu_available() else "cpu")
        full_state_dict.update(shard_state_dict)

    return full_state_dict



# Copyright (c) 2025 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import functools
import math
from abc import abstractmethod

import numpy as np
import torch
from torch.autograd import Function
from torch.nn import functional as F

from neural_compressor.torch.utils import accelerator, logger, set_module

from ..weight_only.modules import HPUWeightOnlyLinear

cast_to_fp8_fcn = lambda x, dtype, scale_inv=None: torch.ops.hpu.cast_to_fp8_v2(x, scale_inv, False, False, dtype)[0]


class HPUMixedPrecisionLinear(HPUWeightOnlyLinear):
    """Weight and Activations quant (W4A8 gptq) Linear for HPU device."""

    def __init__(
        self,
        in_features,
        out_features,
        bias,
        **kwargs,
    ):
        """Init the HPUMixedPrecisionLinear object."""
        super(HPUMixedPrecisionLinear, self).__init__(in_features, out_features, bias=bias)

    def forward(self, input):
        """The forward function of HPUMixedPrecisionLinear."""
        input_dtype = input.dtype
        output_shape = input.shape[:-1] + (self.out_features,)
        scales = self.scales
        qweight = self.qweight
        zeros = self.qzeros
        weight = torch.ops.hpu.convert_from_uint4(qweight, scales, zeros, torch.float8_e4m3fn)
        output = self.matmul_internal(input, weight)
        output = output.to(dtype=input_dtype).reshape(
            output_shape
        )  # A cast is needed here as for some reason the vecquant2matmul_faster_old still allocate a float32 output.
        output = output + self.bias if self.bias is not None else output
        return output

    def forward_mesaure(self, input):
        """The measure forward for w4a8 scheme.

        Measuring is done in bf16 precision matmul for
        generation of fp8 scales for the activations.
        """
        input_dtype = input.dtype
        output_shape = input.shape[:-1] + (self.out_features,)
        scales = self.scales
        scale_bf16_to_fp8 = self.scale_bf16_to_fp8
        qweight = self.qweight
        zeros = self.qzeros
        weight = torch.ops.hpu.convert_from_uint4(qweight, scales, zeros, torch.float8_e4m3fn)
        weight = weight.to(input_dtype) * scale_bf16_to_fp8
        output = self.matmul_internal(input, weight)
        output = output.to(dtype=input_dtype).reshape(output_shape)
        output = output + self.bias if self.bias is not None else output
        return output

    def forward_inference(self, input):
        """Performs the inference-only forward pass.

        The `forward_inference` method differs from `forward` in two key ways:
            1) The `scale_other` parameter in `matmul_internal` is fixed and pre-absorbed into the scales.
            2) The weight is preconverted to FP8 during initialization, whereas in `forward`, it is converted to FP8 on the fly.
        """
        input_dtype = input.dtype
        output_shape = input.shape[:-1] + (self.out_features,)
        scales = self.scales
        qweight = self.qweight
        zeros = self.qzeros

        weight = torch.ops.hpu.convert_from_uint4(
            qweight, scales, zeros, torch.float8_e4m3fn
        )  # todo: div scales in init
        output = self.matmul_internal(input, weight)
        output = output.to(dtype=input_dtype).reshape(output_shape)
        output = output + self.bias if self.bias is not None else output
        return output

    @staticmethod
    def convert_from_weight_only(obj):
        bias = obj.bias is not None
        new_self = HPUMixedPrecisionLinear(obj.in_features, obj.out_features, bias)
        for attr, value in vars(obj).items():
            setattr(new_self, attr, value)
        new_self.matmul_internal.no_input_quant = True  # flag for 8bit input, which shouldn't be quantized in matmul
        if hasattr(new_self, "_original_forward"):
            new_self.forward = new_self._original_forward
        return new_self

    def prepare_from_weight_only(obj):
        # prepare class for measurement
        bias = obj.bias is not None
        new_self = HPUMixedPrecisionLinear(obj.in_features, obj.out_features, bias)
        for attr, value in vars(obj).items():
            setattr(new_self, attr, value)
        new_self._original_forward = new_self.forward
        new_self.forward = new_self.forward_mesaure
        return new_self

    def post_process_for_inference(self):
        """Post process for inference."""
        from neural_compressor.torch.algorithms.fp8_quant._core.quant_dequant import QuantDequantNone, QuantInput
        from neural_compressor.torch.algorithms.fp8_quant._quant_common.helper_modules import PatchedMatmul

        self = self.to("hpu")
        module = self
        patched_matmul: PatchedMatmul = module.matmul_internal
        w_bf16_to_fp8_scale = module.w_bf16_to_fp8_scale
        assert w_bf16_to_fp8_scale.numel() == 1, f"Only support per-tensor scale, but got {w_bf16_to_fp8_scale}"
        w_scales = module.scales
        w_fp8_to_int4_scale = w_scales / w_bf16_to_fp8_scale
        module.scales.data.copy_(w_fp8_to_int4_scale)
        act_scales = module.act_scales
        patched_matmul.scale_input = act_scales.item()
        patched_matmul.scale_other = w_bf16_to_fp8_scale.item()
        input_quantizer: QuantInput = patched_matmul.quant_input_0
        input_quantizer.scale_inv = 1.0 / act_scales.item()
        self.forward = self.forward_inference


def assign_scales_to_patched_matmul_(model):
    for _, module in model.named_modules():
        if isinstance(module, HPUMixedPrecisionLinear):
            logger.debug(f"Assign scales to PatchedMatmul in {module}")
            module.post_process_for_inference()


def _create_fp8_config_from_quant_config(model):
    from neural_compressor.torch.quantization import FP8Config

    blocklist = {"types": [], "names": []}
    allowlist = {"types": ["Matmul"], "names": []}
    fp8_config = FP8Config(mode="LOAD", allowlist=allowlist, blocklist=blocklist, scale_format="scalar")
    return fp8_config


@functools.lru_cache(maxsize=None)
def init_quantized_func_wrapper_factory_once():
    from neural_compressor.torch.algorithms.fp8_quant._core.quantized_func_wrappers import (
        init_quantized_func_wrapper_factory,
    )

    init_quantized_func_wrapper_factory()


def patch_model_for_inference_(model):
    init_quantized_func_wrapper_factory_once()
    from neural_compressor.torch.algorithms.fp8_quant import prep_model as fp8_quant_prep_model_

    fp8_config = _create_fp8_config_from_quant_config(model)
    fp8_config.save_temp_json_file()
    fp8_quant_prep_model_(model, fp8_config.json_file)
    assign_scales_to_patched_matmul_(model)


def replace_hpu_woq_with_hpu_mixed_precision_linear(woq_model):
    for name, module in woq_model.named_modules():
        if isinstance(module, HPUWeightOnlyLinear):
            new_module = HPUMixedPrecisionLinear.convert_from_weight_only(module)
            set_module(woq_model, name, new_module)
    # If we already have the calibration results, we can directly replace the matmul with PatchedMatmul.
    patch_model_for_inference_(woq_model)
    return woq_model



# Copyright (c) 2025 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from enum import Enum, auto

import torch

from neural_compressor.torch.utils.auto_accelerator import auto_detect_accelerator
from .quantized_func_wrapper import QuantizedFuncWrapperFactory


"""
Functions to interact with QuantizedFuncWrapperFactory singleton object
"""

def get_quantized_func_wrapper(op_type, scale_format, is_dynamic=False):
    return QuantizedFuncWrapperFactory.get_quantized_func_wrapper_object(op_type, scale_format, is_dynamic)


def init_quantized_func_wrapper_factory():
    curr_accelerator = auto_detect_accelerator()
    device_name = curr_accelerator.name()
    if device_name == "hpu":
        from .hpu.hpu_quantized_func_wrapper import init_hpu_quantized_func_wrapper_factory
        init_hpu_quantized_func_wrapper_factory()
    elif device_name == "xpu":
        from .xpu.xpu_quantized_func_wrapper import init_xpu_quantized_func_wrapper_factory
        init_xpu_quantized_func_wrapper_factory()
    elif device_name == "cpu":
        from .cpu.cpu_quantized_func_wrapper import init_cpu_quantized_func_wrapper_factory
        init_cpu_quantized_func_wrapper_factory()
    else:
        raise ValueError("Unknown device type - {}".format(device_name))


def clear_quantized_func_wrapper_factory():
    QuantizedFuncWrapperFactory.clear()



# Copyright (c) 2024 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Configuration for HQQ."""

import os
from collections import namedtuple
from dataclasses import dataclass
from typing import Dict, Optional

from typing_extensions import TypeAlias

__all__ = [
    "ConfigMappingType",
    "HQQModuleConfig",
    "QTensorConfig",
    "hqq_global_option",
    "default_hqq_module_config",
    "default_weight_quant_config",
    "default_scale_quant_config",
    "default_zero_quant_config",
]


class HQQGlobalOptions:
    """Global options for HQQ."""

    use_half = os.getenv("HQQ_NOT_USE_HALF", "0") == "0"


hqq_global_option = HQQGlobalOptions()


@dataclass
class QTensorConfig:
    """Configuration class for quantized tensors."""

    nbits: int
    channel_wise: bool = True
    group_size: int = 128
    optimize: bool = True
    round_zero: Optional[bool] = False
    pack: bool = True

    def __repr__(self) -> str:
        """Return a string representation of the QTensorConfig."""
        return (
            f"QTensorConfig(nbits={self.nbits}, channel_wise={self.channel_wise}, "
            f"group_size={self.group_size}, optimize={self.optimize}, "
            f"round_zero={self.round_zero}, pack={self.pack})"
        )


default_weight_quant_config = QTensorConfig(nbits=4, channel_wise=True, group_size=128, optimize=True, round_zero=True)
default_scale_quant_config = QTensorConfig(nbits=8, channel_wise=True, group_size=64, optimize=False, round_zero=None)
default_zero_quant_config = QTensorConfig(nbits=8, channel_wise=False, group_size=None, optimize=False, round_zero=None)


class HQQModuleConfig(
    namedtuple(
        "HQQModuleConfig",
        ["weight", "scale", "zero"],
    )
):
    """Configuration class for HQQModule.

    Args:
        weight (Any): The weight quantization configuration.
        scale (Any): The scale quantization configuration.
        zero (Any): The zero quantization configuration.
    """

    def __new__(
        cls,
        weight=default_weight_quant_config,
        scale=default_scale_quant_config,
        zero=default_zero_quant_config,
    ):
        """Create a new HQQModuleConfig."""
        return super().__new__(cls, weight, scale, zero)

    def __repr__(self) -> str:
        """Return a string representation of the HQQModuleConfig."""
        return (
            f"HQQModuleConfig(\n" f"    weight={self.weight},\n" f"    scale={self.scale},\n" f"    zero={self.zero}\n)"
        )


default_hqq_module_config = HQQModuleConfig(
    weight=default_weight_quant_config,
    scale=default_scale_quant_config,
    zero=default_zero_quant_config,
)


ConfigMappingType: TypeAlias = Dict[str, HQQModuleConfig]



#
# -*- coding: utf-8 -*-
#
# Copyright (c) 2024 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Half-precision Convert for Torch Modules."""

from typing import Dict, Tuple

import torch

from neural_compressor.common import logger
from neural_compressor.torch.algorithms.mixed_precision.module_wrappers import HalfPrecisionModuleWrapper
from neural_compressor.torch.utils import get_accelerator


class HalfPrecisionConverter:
    """Converter Class for FP16 and BF16."""

    dtype_mapping = {
        "fp16": torch.float16,
        "bf16": torch.bfloat16,
    }

    def __init__(self, configs_mapping: Dict[Tuple[str], object], *args, **kwargs):
        """Initialize the Half-precision Converter with config.

        Args:
            configs_mapping (Dict): config class for mixed-precision.
        """
        self.configs_mapping = configs_mapping
        self.device = get_accelerator().current_device_name()

    def convert(self, model: torch.nn.Module):
        """Convert to FP16 or BF16 model.

        Args:
            model (torch.nn.Module): the input model.

        Returns:
            mixed_precision_model (torch.nn.Module): model with mixed-precision.
        """
        if len(self.configs_mapping) > 0:
            logger.info("Convert operators to half-precision")

        if next(model.parameters()).is_cuda:
            self.device = "cuda"
        elif next(model.parameters()).is_cpu:
            self.device = "cpu"

        mixed_precision_model = self._wrap_half_precision_model(model)
        mixed_precision_model.to(self.device)

        return mixed_precision_model

    def _wrap_half_precision_model(self, model: torch.nn.Module, prefix=""):
        """Wrap and replace half-precision target modules.

        Args:
            model (torch.nn.Module): the input module.
            prefix (str): the name prefix for named children.

        Returns:
            model (torch.nn.Module): the model whose target modules have been wrapped.
        """
        for name, child in model.named_children():
            op_name = prefix + "." + name if prefix != "" else name
            for op_info, config in self.configs_mapping.items():
                if op_name == op_info[0] and config.dtype in ("fp16", "bf16"):
                    child = HalfPrecisionModuleWrapper(
                        module=child, device=self.device, dtype=self.dtype_mapping[config.dtype]
                    )
            else:
                self._wrap_half_precision_model(child, op_name)
                setattr(model, name, child)

        return model



# Copyright (c) 2024 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Utility functions for Static quantization."""


import copy
import json
import os
import re
from collections import OrderedDict
from typing import Dict, List, Union

import torch
from packaging.version import Version

try:
    import intel_extension_for_pytorch as ipex
except:  # pragma: no cover
    pass

from neural_compressor.common.utils import DEFAULT_WORKSPACE, CpuInfo
from neural_compressor.torch.utils import Statistics, get_ipex_version, get_torch_version, logger

version = get_torch_version()
ipex_ver = get_ipex_version()
ipex_config_path = os.path.join(DEFAULT_WORKSPACE, "ipex_config_tmp.json")

unify_op_type_mapping_ipex = {
    "Convolution_Relu": "Conv2d",
    "Convolution_Sum_Relu": "Conv2d",
    "Convolution_BatchNorm": "Conv2d",
    "<class 'torch.nn.modules.conv.Conv1d'>": "Conv1d",
    "<class 'torch.nn.modules.conv.Conv2d'>": "Conv2d",
    "<class 'torch.nn.modules.conv.Conv3d'>": "Conv3d",
    "<class 'torch.nn.modules.activation.ReLU'>": "ReLU",
    "<class 'torch.nn.modules.sparse.EmbeddingBag'>": "EmbeddingBag",
    "<method 'add' of 'torch._C._TensorBase' objects>": "add",  # for IPEX < 2.2
    "<method 'add' of 'torch._C.TensorBase' objects>": "add",  # for IPEX >= 2.2
    "<class 'torch.nn.modules.pooling.AdaptiveAvgPool2d'>": "AdaptiveAvgPool2d",
    "Linear_Relu": "Linear",
    "Linear_add": "Linear",
    "<class 'torch.nn.modules.linear.Linear'>": "Linear",
    "<class 'torch.nn.modules.pooling.MaxPool2d'>": "MaxPool2d",
    "re": {
        "<built-in method matmul of type object at": "matmul",
        "<built-in method add of type object at": "add",
        "<built-in method bmm of type object at": "bmm",
    },
}

BLOCK_PATTERNS = [
    # [['OP_TYPE1', NUM_OPS], ['OP_TYPE2', NUM_OPS], ...]
    [["Linear", 4], ["Linear", 4]],  # TODO add model name
    [["Linear", 2], ["Linear", 2]],  # TODO add model name
    [["Conv1D", 2], ["Conv1D", 2]],  # GPT-2
    [["Linear", 4], ["Linear", 3]],  # Llama
    [["Linear", 4], ["Linear", 2]],  # T5-Encoder, OPT
    [["Linear", 4], ["Linear", 1], ["Linear", 1]],  # Bert
    [["Linear", 4], ["Linear", 4], ["Linear", 2]],  # T5-Decoder
]


def cfg_to_qconfig(tune_cfg, cfgs, op_infos_from_cfgs, output_tensor_id_op_name):  # pragma: no cover
    """Updates json file in ipex_config_path.

    Args:
        tune_cfg (dict): dictionary of quantization configuration.
        cfgs (dict): configs loaded from ipex config path.
        op_infos_from_cfgs (dict): dict containing configs that have been parsed for each op.
        output_tensor_ids_op_name (dict): dict containing op names corresponding to 'op_infos_from_cfgs'.

    Returns:
        user_cfg (dict): quantization configuration for ops.
    """
    assert cfgs is not None, "No configure for IPEX int8 model..."
    op_infos = copy.deepcopy(op_infos_from_cfgs)
    cfgs, user_cfg = check_cfg_and_qconfig(tune_cfg["op"], cfgs, op_infos, output_tensor_id_op_name)
    with open(ipex_config_path, "w") as write_f:
        json.dump(cfgs, write_f, indent=4)
    return user_cfg


def check_cfg_and_qconfig(user_cfg, cfgs, op_infos_from_cfgs, output_tensor_ids_op_name):  # pragma: no cover
    """Check configs and quantization configs.

    Args:
        user_cfg (dict): quantization configuration for ops.
        cfgs (dict): configs loaded from ipex config path.
        op_infos_from_cfgs (dict): dict containing configs that have been parsed for each op.
        output_tensor_ids_op_name (dict): dict containing op names corresponding to 'op_infos_from_cfgs'.

    Returns:
        cfgs (dict): updated configs.
    """
    ori_user_cfg = copy.deepcopy(user_cfg)
    tmp_user_cfg = OrderedDict()
    for op in user_cfg:  # map ipex op_name to pt op_name
        for i, op_name in enumerate(op):
            for ops, _ in op_infos_from_cfgs.items():
                if "fqn" in op_infos_from_cfgs[ops].keys() and op_infos_from_cfgs[ops]["fqn"] == op_name:
                    if op_infos_from_cfgs[ops]["op_type"] in unify_op_type_mapping_ipex:
                        ori_op = (tuple(ops), unify_op_type_mapping_ipex[op_infos_from_cfgs[ops]["op_type"]])
                        tmp_user_cfg[((ori_op[0],), ori_op[1])] = user_cfg[op]
                        break

    for op_name in tmp_user_cfg:
        inc_op_cfg = tmp_user_cfg[op_name]
        for i, name in enumerate(op_name[0]):
            # to int8
            ipex_op_cfg = op_infos_from_cfgs[name]
            input_tensor_infos = ipex_op_cfg["input_tensor_infos"]
            if op_name[1] == "Linear" or op_name[1] == "Linear&add":  # record op_name for possible op-wise fallback
                logger.debug(f"ipex_op_cfg['fqn'] - op_name {ipex_op_cfg['fqn']}  {op_name}")
            for index, input_tensor_info in enumerate(input_tensor_infos):
                if "force_dtype" not in input_tensor_info.keys():
                    continue
                if (
                    input_tensor_info["force_dtype"] == "torch.qint8"
                    or input_tensor_info["force_dtype"] == "torch.quint8"
                ):
                    # int8 -> int8
                    if inc_op_cfg["weight"]["dtype"] == "int8":
                        inc_scheme = inc_op_cfg["activation"]["scheme"]
                        inc_algorithm = inc_op_cfg["activation"]["algorithm"]
                        ipex_op_cfg["input_tensor_infos"] = input_tensor_infos
                        if (
                            "op_type" in ipex_op_cfg
                            and ipex_op_cfg["op_type"] == "<class 'torch.nn.modules.linear.Linear'>"
                        ):
                            smooth_quant_enable = True
                        else:
                            smooth_quant_enable = False
                        activation_observer = generate_activation_observer(
                            inc_scheme, inc_algorithm, smooth_quant=False, smooth_quant_enable=smooth_quant_enable
                        )
                        if inc_scheme == "sym":
                            input_tensor_infos[index]["force_dtype"] = "torch.qint8"
                        if inc_scheme == "asym":
                            input_tensor_infos[index]["force_dtype"] = "torch.quint8"
                        ipex_op_cfg["activation_observer"] = activation_observer
                    # int8 -> fp32
                    else:
                        input_tensor_infos[index]["force_dtype"] = "torch.float32"
                    # modify pre_op output inf_dtype
                    if i == 0:
                        input_tensor_id = input_tensor_info["id"]
                        input_tensor_dtype = input_tensor_info["force_dtype"]
                        if input_tensor_id in output_tensor_ids_op_name.keys():
                            pre_op_name = output_tensor_ids_op_name[input_tensor_id]
                            pre_op_module = pre_op_name[0][0]
                            pre_op_state = pre_op_name[0][1]
                            pre_op_index = pre_op_name[0][2]
                            pre_op_infos = cfgs[pre_op_module][pre_op_state][pre_op_index]
                            pre_op_output_infos = pre_op_infos["output_tensor_infos"]
                            for index, pre_op_output in enumerate(pre_op_output_infos):
                                if pre_op_output["id"] == input_tensor_id:
                                    pre_op_output_infos[index]["inf_dtype"] = input_tensor_dtype
                                else:
                                    pass
                            pre_op_infos["output_tensor_infos"] = pre_op_output_infos
                            cfgs[pre_op_module][pre_op_state][pre_op_index] = pre_op_infos
                        else:
                            pass
            cfgs[name[0]][name[1]][name[2]] = ipex_op_cfg
    return cfgs, ori_user_cfg


def generate_xpu_qconfig(tune_cfg):  # pragma: no cover
    """Generates qconfig for quantiztaion on xpu device.

    Args:
        tune_cfg (dict): dictionary of quantization configuration.

    Returns:
        qconfig (dict): quantization configuration for ops.
    """
    # qconfig observer & config constants for ipex-xpu
    from torch.ao.quantization import HistogramObserver, MinMaxObserver, QConfig

    act_observer_minmax_asym = MinMaxObserver.with_args(quant_min=0, quant_max=127)
    act_observer_minmax_sym = MinMaxObserver.with_args(
        dtype=torch.qint8, qscheme=torch.per_tensor_symmetric, quant_min=-128, quant_max=127
    )
    act_observer_kl_asym = HistogramObserver.with_args(quant_min=0, quant_max=127)
    act_observer_kl_sym = HistogramObserver.with_args(
        dtype=torch.qint8, qscheme=torch.per_tensor_symmetric, quant_min=-128, quant_max=127
    )
    # no tuning for granularity due to tuning space
    weight_observer_minmax_sym = MinMaxObserver.with_args(dtype=torch.qint8, qscheme=torch.per_tensor_symmetric)

    qconfig = {}
    user_cfg = copy.deepcopy(tune_cfg["op"])
    for _, cfg in user_cfg.items():
        act_algo = cfg["activation"]["algorithm"]
        act_sym = cfg["activation"]["scheme"]
        break

    if act_algo == "minmax":
        if act_sym == "sym":
            activation = act_observer_minmax_sym
        else:
            activation = act_observer_minmax_asym
    else:
        if act_sym == "sym":
            activation = act_observer_kl_sym
        else:
            activation = act_observer_kl_asym

    qconfig[""] = QConfig(activation=activation, weight=weight_observer_minmax_sym)

    for (op_name, op_type), cfg in user_cfg.items():
        if cfg["weight"]["dtype"] == "fp32":
            qconfig[op_name] = None
    return qconfig


def generate_activation_observer(
    scheme, algorithm, smooth_quant=False, smooth_quant_enable=False, alpha=0.5
):  # pragma: no cover
    """This is a helper method to generate an activation observer.

    Args:
        scheme (str): Quantization scheme to be used.
        algorithm (str): What algorithm for computing the quantization parameters based on.

    Returns:
        An observer.
    """
    kl_activation_observer = {
        "name": "HistogramObserver",
        "bins": 2048,
        "dtype": "torch.quint8",
        "qscheme": "torch.per_tensor_affine",
        "reduce_range": False,
        "quant_min": 0,
        "quant_max": 255,
    }
    minmax_activation_observer = {
        "name": "MinMaxObserver",
        "dtype": "torch.quint8",
        "qscheme": "torch.per_tensor_affine",
        "reduce_range": False,
        "quant_min": 0,
        "quant_max": 255,
    }
    smoothquant_kl_activation_observer = {
        "name": "SmoothQuantActivationObserver",
        "smooth_quant_enabled": smooth_quant_enable,
        "dtype": "torch.quint8",
        "qscheme": "torch.per_tensor_affine",
        "reduce_range": False,
        "quant_min": 0,
        "quant_max": 255,
        "alpha": 0.5 if alpha == "auto" else alpha,
        "act_observer": kl_activation_observer,
        "act_ic_observer": {
            "name": "PerChannelMinMaxObserver",
            "ch_axis": -1,
            "dtype": "torch.quint8",
            "qscheme": "torch.per_channel_affine",
            "reduce_range": False,
            "quant_min": 0,
            "quant_max": 255,
        },
    }
    smoothquant_minmax_activation_observer = {
        "name": "SmoothQuantActivationObserver",
        "smooth_quant_enabled": smooth_quant_enable,
        "dtype": "torch.quint8",
        "qscheme": "torch.per_tensor_affine",
        "reduce_range": False,
        "quant_min": 0,
        "quant_max": 255,
        "alpha": 0.5 if alpha == "auto" else alpha,
        "act_observer": minmax_activation_observer,
        "act_ic_observer": {
            "name": "PerChannelMinMaxObserver",
            "ch_axis": -1,
            "dtype": "torch.quint8",
            "qscheme": "torch.per_channel_affine",
            "reduce_range": False,
            "quant_min": 0,
            "quant_max": 255,
        },
    }
    REDUCE_RANGE = False if CpuInfo().vnni else True
    if REDUCE_RANGE:
        minmax_activation_observer["reduce_range"] = REDUCE_RANGE
        kl_activation_observer["reduce_range"] = REDUCE_RANGE
    if scheme == "sym":
        minmax_activation_observer["qscheme"] = "torch.per_tensor_symmetric"
        minmax_activation_observer["dtype"] = "torch.qint8"
        minmax_activation_observer["quant_min"] = -128
        minmax_activation_observer["quant_max"] = 127
        kl_activation_observer["qscheme"] = "torch.per_tensor_symmetric"
        kl_activation_observer["dtype"] = "torch.qint8"
        kl_activation_observer["quant_min"] = -128
        kl_activation_observer["quant_max"] = 127
    if smooth_quant and smooth_quant_enable:
        if algorithm == "kl":
            return smoothquant_kl_activation_observer
        if algorithm == "minmax":
            return smoothquant_minmax_activation_observer
    else:
        if algorithm == "kl":
            return kl_activation_observer
        if algorithm == "minmax":
            return minmax_activation_observer


def get_quantizable_ops_recursively(model, example_inputs):  # pragma: no cover
    """Get all quantizable ops from model.

    Args:
        model (object): input model
        example_inputs (dict|list|tuple|torch.Tensor): used to trace torch model.

    Returns:
        quantizable_ops (list): list of tuples of op_name and op_type.
        cfgs (dict): dict of configuration.
        op_infos_from_cfgs (dict): dict containing configs that have been parsed for each op.
        output_tensor_ids_op_name (dict): dict containing op names corresponding to 'op_infos_from_cfgs'.
    """
    quantizable_ops = []
    op_name_info = []
    # group ops by position for transform-based model
    detector = TransformerBasedModelBlockPatternDetector(model)
    detect_result = detector.detect_block()
    attention_block = detect_result.get("attention_blocks", None)
    ffn_blocks = detect_result.get("ffn_blocks", None)
    logger.info(f"Attention Blocks: {len(attention_block)}")
    logger.info(f"FFN Blocks: {len(ffn_blocks)}")
    if not os.path.exists(ipex_config_path):
        assert isinstance(model, torch.nn.Module), "The model passed in is not the instance of torch.nn.Module"

    if hasattr(model, "save_qconf_summary"):  # pragma: no cover
        os.makedirs(os.path.dirname(ipex_config_path), exist_ok=True)
        model.save_qconf_summary(qconf_summary=ipex_config_path)
    else:
        model.eval()

        # create a quantization config file for intel pytorch extension model
        os.makedirs(os.path.dirname(ipex_config_path), exist_ok=True)
        assert example_inputs is not None, "IPEX need q_dataloader or example_inputs to prepare the model"
        from torch.ao.quantization import MinMaxObserver, PerChannelMinMaxObserver, QConfig

        if ipex_ver.release >= Version("2.1").release:
            # HistogramObserver will cause a performance issue.
            # static_qconfig = ipex.quantization.default_static_qconfig_mapping
            qconfig = QConfig(
                activation=MinMaxObserver.with_args(qscheme=torch.per_tensor_affine, dtype=torch.quint8),
                weight=PerChannelMinMaxObserver.with_args(dtype=torch.qint8, qscheme=torch.per_channel_symmetric),
            )
            from torch.ao.quantization import QConfigMapping

            static_qconfig = QConfigMapping().set_global(qconfig)
        else:
            static_qconfig = QConfig(
                activation=MinMaxObserver.with_args(qscheme=torch.per_tensor_affine, dtype=torch.quint8),
                weight=PerChannelMinMaxObserver.with_args(dtype=torch.qint8, qscheme=torch.per_channel_symmetric),
            )

        if isinstance(example_inputs, dict):
            model = ipex.quantization.prepare(model, static_qconfig, example_kwarg_inputs=example_inputs, inplace=True)
        else:
            model = ipex.quantization.prepare(model, static_qconfig, example_inputs=example_inputs, inplace=True)
        simple_inference(model, example_inputs, iterations=1)
        model.save_qconf_summary(qconf_summary=ipex_config_path)

    map_op_name_to_fqn = {}
    with open(ipex_config_path, "r") as f:
        cfgs = json.load(f)
        (
            ops_name,
            op_infos_from_cfgs,
            input_tensor_id_op_name,
            output_tensor_id_op_name,
        ) = parse_cfgs(cfgs)
        quantizable_op_names = get_quantizable_ops_from_cfgs(ops_name, op_infos_from_cfgs, input_tensor_id_op_name)
        for name in quantizable_op_names:
            # name : list
            if len(name) == 1:
                module_key = name[0][0]
                op_cfg_id = name[0][2]
                ipex_op_type = cfgs[module_key]["q_op_infos"][op_cfg_id]["op_type"]
                module_fqn = cfgs[module_key]["q_op_infos"][op_cfg_id].get("fqn", None)

                if ipex_op_type in unify_op_type_mapping_ipex:
                    quantizable_ops.append((tuple(name), unify_op_type_mapping_ipex[ipex_op_type]))
                    map_op_name_to_fqn[(tuple(name), ipex_op_type)] = module_fqn
                    if "class" in ipex_op_type:  # "<class 'torch.nn.modules.activation.ReLU'>"
                        _, class_name = ipex_op_type.strip(">").split("'")[1].rsplit(".", 1)
                        op_name_info.append((module_fqn, class_name))
                    elif "method" in ipex_op_type:  # "<method 'add' of 'torch._C._TensorBase' objects>"
                        method = ipex_op_type.split("'")[1]
                        op_name_info.append((module_fqn, method))
                    elif "_" in ipex_op_type:  # "Convolution_Relu", "Linear_Relu"
                        op_name_info.append((module_fqn, ipex_op_type.split("_")[0]))
                else:
                    re_flag = False
                    for pattern, unify_op_type in unify_op_type_mapping_ipex["re"].items():
                        if re.match(pattern, ipex_op_type):
                            re_flag = True
                            quantizable_ops.append((tuple(name), unify_op_type))
                            map_op_name_to_fqn[(tuple(name), unify_op_type)] = module_fqn
                            op_name_info.append((module_fqn, ipex_op_type))
                            break
                    if not re_flag:
                        quantizable_ops.append((tuple(name), ipex_op_type))
                        map_op_name_to_fqn[(tuple(name), ipex_op_type)] = module_fqn
                        op_name_info.append((module_fqn, ipex_op_type))
            else:
                op_type = ""
                for op_name in name:
                    module_key = op_name[0]
                    op_cfg_id = op_name[2]
                    single_op_type = cfgs[module_key]["q_op_infos"][op_cfg_id]["op_type"]
                    if single_op_type in unify_op_type_mapping_ipex:
                        single_op_type = unify_op_type_mapping_ipex[single_op_type]
                    op_type += "&" + single_op_type if op_type else single_op_type
                quantizable_ops.append((tuple(name), op_type))
                _module_key = name[0][0]
                _op_cfg_id = name[0][2]
                module_fqn = cfgs[_module_key]["q_op_infos"][_op_cfg_id]["fqn"]
                map_op_name_to_fqn[(tuple(name), op_type)] = module_fqn
                op_name_info.append((module_fqn, op_type))

    logger.debug("Map op name to fqn: ")
    logger.debug(map_op_name_to_fqn)
    logger.info("Attention Blocks : ")
    logger.info(attention_block)
    logger.info("FFN Blocks : ")
    logger.info(ffn_blocks)
    return quantizable_ops, cfgs, op_infos_from_cfgs, output_tensor_id_op_name, op_name_info


def simple_inference(q_model, example_inputs, iterations=1):
    """The function is used for ipex warm-up inference."""
    for _ in range(iterations):
        if isinstance(example_inputs, tuple) or isinstance(example_inputs, list):
            q_model(*example_inputs)
        elif isinstance(example_inputs, dict):
            q_model(**example_inputs)
        else:
            q_model(example_inputs)


def dump_model_op_stats(user_cfg):
    """This is a function to dump quantizable ops of model to user.

    Args:
        user_cfg (dict): quantization config

    Returns:
        None
    """
    res = dict()
    for k, v in user_cfg.items():
        op_type = k[1]
        if op_type not in res.keys():
            res[op_type] = {"INT8": 0, "BF16": 0, "FP32": 0}
        if v["weight"]["dtype"] == "int8":
            res[op_type]["INT8"] += 1
        elif v["weight"]["dtype"] == "fp32":
            res[op_type]["FP32"] += 1

    output_data = [
        [op_type, sum(res[op_type].values()), res[op_type]["INT8"], res[op_type]["BF16"], res[op_type]["FP32"]]
        for op_type in res.keys()
    ]

    Statistics(
        output_data, header="Mixed Precision Statistics", field_names=["Op Type", "Total", "INT8", "BF16", "FP32"]
    ).print_stat()


def get_depth(d) -> int:
    """Query the depth of the dict."""
    if isinstance(d, dict):
        return 1 + max(get_depth(v) for v in d.values())
    return 0


def get_dict_at_depth(d, target_depth, result, depth=0):
    """Get all sub-dicts that are at a specified depth in a nested dict."""
    if depth == target_depth:
        result.append(d)
        return
    elif depth < target_depth and isinstance(d, dict):
        for k, v in d.items():
            get_dict_at_depth(v, target_depth, result, depth=depth + 1)


def get_element_under_depth(d, ops_lst):
    """Get all values in a nested dict."""
    if isinstance(d, dict):
        for k, v in d.items():
            get_element_under_depth(v, ops_lst)
    else:
        ops_lst.append(d)


def parse_cfgs(cfgs):  # pragma: no cover
    """Parse configs.

    Args:
        cfgs (dict): the input configs.

    Returns:
        ops_name (list): list of op names.
        tune_cfg (dict): dictionary of quantization configuration.
        op_infos_from_cfgs (dict): op infos from configs.
        output_tensor_ids_op_name (dict): dictionary of output tensor op names.
    """
    ops_name = []
    layer_output_infos_ids = []
    op_infos_from_cfgs = {}
    # record input_tensor_id and op_name
    # {"0": [(" ", "q_op_infos", "0"), (" ", "q_op_infos", "1")]}
    input_tensor_ids_op_name = {}
    output_tensor_ids_op_name = {}
    for module_key in cfgs.keys():
        for state in cfgs[module_key]:
            if state == "layer_output_infos":
                for index, op_info in enumerate(cfgs[module_key][state]):
                    name = (module_key, state, index)
                    ops_name.append(name)
                    layer_output_infos_ids.append(op_info["id"])
                    op_infos_from_cfgs[name] = op_info
                continue
            for op_cfg_id in cfgs[module_key][state].keys():
                op_info = cfgs[module_key][state][op_cfg_id]
                name = (module_key, state, op_cfg_id)
                if name not in ops_name:
                    ops_name.append(name)
                else:
                    assert False, "Please check IPEX int8 configure json whether have the same name ops"
                op_infos_from_cfgs[name] = op_info
                input_tensors = op_info["input_tensor_infos"]
                for input_tensor in input_tensors:
                    if "id" not in input_tensor.keys():
                        continue
                    else:
                        input_tensor_id = input_tensor["id"]
                    if input_tensor_id not in input_tensor_ids_op_name.keys():
                        input_tensor_ids_op_name[input_tensor_id] = [name]
                    else:
                        input_tensor_ids_op_name[input_tensor_id].append(name)
                output_tensors = op_info["output_tensor_infos"]
                for output_tensor in output_tensors:
                    if "id" not in output_tensor.keys():
                        continue
                    else:
                        output_tensor_id = output_tensor["id"]
                    if output_tensor_id not in output_tensor_ids_op_name.keys():
                        output_tensor_ids_op_name[output_tensor_id] = [name]
                    else:
                        output_tensor_ids_op_name[output_tensor_id].append(name)
    return ops_name, op_infos_from_cfgs, input_tensor_ids_op_name, output_tensor_ids_op_name


def get_quantizable_ops_from_cfgs(ops_name, op_infos_from_cfgs, input_tensor_ids_op_name):  # pragma: no cover
    """Get quantizable ops from configs, combine fused ops as one op.

    Args:
        ops_name (list): list of op names.
        op_infos_from_cfgs (dict): op infos from configs.
        input_tensor_ids_op_name (dict): dictionary of input tensor op names.

    Returns:
        cfgs (dict).
    """
    quantizable_ops = []
    seen_ops = []
    for name in ops_name:
        start = True
        if name in seen_ops:
            continue
        elif name[1] not in ["q_op_infos"]:
            continue
        else:
            # judge fuse ops the first op
            op_info = op_infos_from_cfgs[name]
            output_tensors = op_info["output_tensor_infos"]
            input_tensors = op_info["input_tensor_infos"]
            start = any(
                [
                    input_tensor["inf_dtype"] != "torch.float32"
                    for input_tensor in input_tensors
                    if "inf_dtype" in input_tensor.keys()
                ]
            )
            if not start:
                continue
            # add quantizable ops, include op and fuse ops.
            q_ops, stack = [], [(name, [])]
            while stack:
                cur_name, cur = stack.pop()
                seen_ops.append(cur_name)
                if cur_name[1] not in ["q_op_infos"]:
                    q_ops.append(cur)
                    break
                op_info = op_infos_from_cfgs[cur_name]
                output_tensors = op_info["output_tensor_infos"]
                for output_tensor in output_tensors:
                    if output_tensor["inf_dtype"] == "torch.qint8" or output_tensor["inf_dtype"] == "torch.quint8":
                        q_ops.append(cur + [cur_name])
                        break
                    try:
                        next_op_names = input_tensor_ids_op_name[output_tensor["id"]]
                        for next_op_name in next_op_names:
                            stack.append((next_op_name, cur + [cur_name]))
                    except:
                        next_op_name = None
                    if next_op_name is None:
                        q_ops.append(cur + [cur_name])
            for q_op in q_ops:
                quantizable_ops.append(q_op)
    return quantizable_ops


class TransformerBasedModelBlockPatternDetector:  # pragma: no cover
    """Detect the attention block and FFN block in transformer-based model."""

    def __init__(self, model: torch.nn.Module, pattern_lst: List[List[Union[str, int]]] = BLOCK_PATTERNS) -> None:
        """Init the block detector.

        Args:
            model: the model to be detected.
            pattern_lst: block pattern list.
        """
        self.model = model
        self.pattern_lst = pattern_lst
        self.pos_info = None

    def detect_block(self) -> Dict[str, List[List[str]]]:
        """Traverse the model definition and return the attention blocks and ffn blocks.

        Returns:
            blocks: A dict include the detected attention blocks and ffn blocks.
        """
        # Step 1: Traverse model definition and record the op position
        if not self.pos_info:
            pos_info = {0: {}}
            self.traverse_model(self.model, result=pos_info)
            self.pos_info = pos_info
        # Step 2: Traverse all blocks in different depths and record the blocks that matched the pattern
        detect_result = []
        for pattern in self.pattern_lst:
            _, result = self._search_pattern(pos_info, pattern)
            if result:
                detect_result.append((result, pattern))
        # Step 3: Get the attention blocks and ffn blocks
        blocks = {"attention_blocks": None, "ffn_blocks": None}
        blocks["attention_blocks"], blocks["ffn_blocks"] = self._group_block(detect_result)
        logger.debug(f'FFN BLOCKS: {blocks["ffn_blocks"]}')
        logger.debug(f'Attention BLOCKS: {blocks["attention_blocks"]}')
        return blocks

    @staticmethod
    def traverse_model(model, prefix="", depth=1, result=None, key=0):
        """Traverse the pytorch model according to its hierarchical structure.

        Args:
            model: input model to be traversed.
            prefix: prefix of module. Defaults to "".
            depth: current traverse depth. Defaults to 1.
            result: depth and its included ops. Defaults to {0: {}}.
            key: current root key. Defaults to 0.
        """
        module_lst = list(model.named_children())
        if len(module_lst) == 0:
            # layer name: 'encoder.layer.7.attention.self.query'
            # model repr: Linear(in_features=768, out_features=768, bias=True)
            # class name: 'Linear'
            result[key] = (prefix, model, model.__class__.__name__)
        for i, (name, sub_module) in enumerate(module_lst, 1):
            indent = "    " * depth
            new_name = prefix + "." + name if prefix != "" else name
            model_type = sub_module.__class__.__name__
            logger.debug(f"Depth: [{depth}]" + indent + f"[{model_type}]{ new_name}")
            sub_key = (depth, i, model_type)
            if sub_key not in result[key]:
                result[key][sub_key] = dict()
            TransformerBasedModelBlockPatternDetector.traverse_model(
                sub_module, prefix=new_name, depth=depth + 1, result=result[key], key=sub_key
            )

    @staticmethod
    def _search_pattern(pos_info: Dict, pattern: List[List[Union[str, int]]]) -> List[List[str]]:
        """Search all blocks that matched the pattern.

        Args:
            pos_info: the position information of ops.
            pattern: block pattern.

        Returns:
            The number of matched blocks and the matched blocks.
        """
        max_depth = get_depth(pos_info)
        matched_cnt = 0
        result = []
        for depth in range(max_depth, -1, -1):
            attention_depth = depth
            depth_block_lst = []
            get_dict_at_depth(pos_info, attention_depth, depth_block_lst, 0)
            target_op_types = set(pair[0] for pair in pattern)
            for i, block in enumerate(depth_block_lst):
                sub_block_lst = []
                get_dict_at_depth(block, 1, sub_block_lst, 0)
                block_pattern = []
                block_result = []
                for sub_block in sub_block_lst:
                    ops_lst = []
                    get_element_under_depth(sub_block, ops_lst)
                    filter_ops = [op for op in ops_lst if op[2] in target_op_types]
                    if len(filter_ops) > 0:
                        sub_block_pattern = [filter_ops[0][2], len(filter_ops)]
                        block_pattern.append(sub_block_pattern)
                        ops_name = [op[0] for op in filter_ops]
                        block_result.append(ops_name)
                if block_pattern == pattern:
                    matched_cnt += 1
                    logger.debug(f"[DEPTH] {depth} [BLOCK] {i},  Found block match pattern {pattern}!!")
                    logger.debug(f"[Block keys] {block.keys()}")
                    logger.debug(f"[Block Ops] { [pair[0] for pair in ops_lst if pair[2] in target_op_types]}")
                    result.append(block_result)
        if matched_cnt > 0:
            logger.info(f" Found {matched_cnt} blocks")
        return matched_cnt, result

    @staticmethod
    def _group_block(detect_result):
        """Collect attention and ffn blocks from detect result."""
        import itertools

        ffn_block_lst = []
        attention_block_lst = []
        for block_lst, pattern in detect_result:
            for block in block_lst:
                # Group the first block as attention blocks and
                # the remaining blocks belong to ffn block.
                if block:
                    attention_block_lst.append(block[0])
                    ffn_block = list(itertools.chain(*block[1:]))
                    if ffn_block:
                        ffn_block_lst.append(ffn_block)
        return attention_block_lst, ffn_block_lst
